/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002, 2004, 2005                                  */
/*                                                                           */
/* This software is licensed under the Eclipse Public License (EPL) V1.0.    */
/*****************************************************************************/

package com.ibm.staf.service.cron;

import com.ibm.staf.*;
import com.ibm.staf.service.*;
import java.util.*;
import java.io.*;
import org.python.util.PythonInterpreter;
import org.python.core.*;

public class CronService implements STAFServiceInterfaceLevel30, Runnable
{
    private STAFHandle fHandle;
    private final String kVersion = "3.3.3";

    // Version of STAF (or later) required for this service
    private final String kRequiredSTAFVersion = "3.1.0";

    private STAFCommandParser fParmsParser = new STAFCommandParser();
    private STAFCommandParser fRegisterParser;
    private STAFCommandParser fListParser;
    private STAFCommandParser fTriggerParser;
    private STAFCommandParser fEnableParser;
    private STAFCommandParser fDisableParser;
    private STAFCommandParser fVersionParser;
    private STAFMapClassDefinition fCronIDMapClass;
    private STAFMapClassDefinition fCronIDLongMapClass;
    private STAFMapClassDefinition fCronIDShortMapClass;
    private STAFMapClassDefinition fTriggerMapClass;
    private Hashtable fRegTable;
    private static boolean DEBUG = true;
    private boolean fOldVarResolution = false;
    private Thread fQueueThread;
    private STAFCommandParser fUnregisterParser;
    private int fCronID = 1;
    private String fLineSep;
    private String fCronServiceMachine = "";
    private String fServiceName = "";
    private String fJythonVersion = "";
    private PythonInterpreter fPythonInterpreter;
    private String fHashtableFileDirectory;
    private String fHashtableFileName;
    private String fLocalMachineName;
    private final String kMinute = "minute";
    private final String kHour = "hour";
    private final String kDay = "day";
    private final String kMonth = "month";
    private final String kWeekday = "weekday";
    private HashMap fSubmittedRequests = new HashMap();
    private HashMap fRequestsAndIDs = new HashMap();
      // key=<machine>:<reqNum>, value=ID
    private HashMap fProcessHandlesAndRequests = new HashMap();
      // key=<machine>:<handle>, value=<machine>:<reqNum>
    static final String sQueueTypeEnd = "STAF/Service/Cron/End";
    private String fServiceHandleName = "";

    public static final int PYTHONERROR = 4001;
    private static final String PYTHONERRORInfo = "Python error";
    private static final String PYTHONERRORDesc =
        "A Python error occurred while evaluating the request to be submitted.";

    public static final int REQUESTNOTSUBMITTED = 4002;
    private static final String REQUESTNOTSUBMITTEDInfo =
        "Request not submitted";
    private static final String REQUESTNOTSUBMITTEDDesc =
        "The request was not submitted because the value of " +
        "STAFCronSubmit was not set to true.";

    public CronService()
    {
    }

    public STAFResult init(STAFServiceInterfaceLevel30.InitInfo info)
    {
        int rc = STAFResult.Ok;

        try
        {
            STAFResult res = new STAFResult();
            fServiceHandleName = "STAF/SERVICE/" + info.name;
            fHandle = new STAFHandle(fServiceHandleName);
            fServiceName = info.name;

            // Resolve the machine name variable for the local machine

            res = STAFUtil.resolveInitVar("{STAF/Config/Machine}", fHandle);

            if (res.rc != STAFResult.Ok) return res;

            fLocalMachineName = res.result;
            
            // Resolve the line separator variable for the local machine

            res = STAFUtil.resolveInitVar("{STAF/Config/Sep/Line}", fHandle);

            if (res.rc != STAFResult.Ok) return res;

            fLineSep = res.result;

            // Verify that the required version of STAF is running on the
            // local service machine.  
            // Note:  Method compareSTAFVersion was added in STAF V3.1.0

            try
            {
                res = STAFUtil.compareSTAFVersion(
                    "local", fHandle, kRequiredSTAFVersion);

                if (res.rc != STAFResult.Ok)
                {
                    if (res.rc == STAFResult.InvalidSTAFVersion)
                    {
                        return new STAFResult(
                            STAFResult.ServiceConfigurationError,
                            "Minimum required STAF version for this service " +
                            "is not running." + fLineSep + res.result);
                    }
                    else
                    {
                        return new STAFResult(
                            STAFResult.ServiceConfigurationError,
                            "Error verifying the STAF version. RC: " + res.rc +
                            ", Additional info: " + res.result);
                    }
                }
            }
            catch (Error err)
            {
                return new STAFResult(
                    STAFResult.ServiceConfigurationError,
                    "This service requires STAF Version " +
                    kRequiredSTAFVersion + " or later."); 
            }

            if (info.serviceJar != null)
            {
                // Create the Jython Library directory if it doesn't exist yet
                // from the jython library data in the service's jarfile.
                // Set the classpath to point to the service's jar file and to
                // the jython.jar.

                res = STAFServiceSharedJython.setupJython(
                    info.serviceJar, fHandle, info.writeLocation);

                if (res.rc != 0)
                {
                    System.out.println(res.result);
                    return res;
                }

                // Get the version of Jython distributed with the Cron service
                    
                fJythonVersion = STAFServiceSharedJython.getJythonVersion();

                // Initialize the PythonInterpreter's python.home to point to the
                // shared jython installation.

                Properties p = new Properties();
                p.setProperty("python.home",
                              STAFServiceSharedJython.getJythonDirName());
                PythonInterpreter.initialize(System.getProperties(), p, null);
            }
            
            // Parse PARMS if provided in the service configuration line

            if (info.parms != null)
            {
                fParmsParser.addOption("OLDVARRESOLUTION", 1,
                                       STAFCommandParser.VALUENOTALLOWED);

                res = handleParms(info);

                if (res.rc != STAFResult.Ok)
                {
                    return new STAFResult(
                        STAFResult.ServiceConfigurationError,
                        "Error validating parameters: RC=" + res.rc +
                        ", Result=" + res.result);
                }
            }

            // instantiate parsers as not case sensitive
            fRegisterParser = new STAFCommandParser(0, false);
            fListParser = new STAFCommandParser(0, false);
            fTriggerParser = new STAFCommandParser(0, false);
            fEnableParser = new STAFCommandParser(0, false);
            fDisableParser = new STAFCommandParser(0, false);
            fUnregisterParser = new STAFCommandParser(0, false);
            fVersionParser = new STAFCommandParser(0, false);

            fRegisterParser.addOption("REGISTER", 1,
                                      STAFCommandParser.VALUENOTALLOWED);

            fRegisterParser.addOption("DESCRIPTION", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("MACHINE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("PYTHONMACHINE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOptionGroup("MACHINE PYTHONMACHINE", 1, 1);

            fRegisterParser.addOption("SERVICE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("PYTHONSERVICE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOptionGroup("SERVICE PYTHONSERVICE", 1, 1);

            fRegisterParser.addOption("REQUEST", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("PYTHONREQUEST", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOptionGroup("REQUEST PYTHONREQUEST", 1, 1);

            fRegisterParser.addOption("PREPARE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("MINUTE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("HOUR", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("DAY", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("MONTH", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("WEEKDAY", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("ONCE", 1,
                                      STAFCommandParser.VALUENOTALLOWED);

            fRegisterParser.addOption("ENABLED", 0,
                                      STAFCommandParser.VALUENOTALLOWED);

            fRegisterParser.addOption("DISABLED", 0,
                                      STAFCommandParser.VALUENOTALLOWED);

            fRegisterParser.addOptionGroup("ENABLED DISABLED", 0, 1);

            // Require at least one time interval option to be specified
            fRegisterParser.addOptionGroup(
                "MINUTE HOUR DAY MONTH WEEKDAY", 1, 5);

            fUnregisterParser.addOption("UNREGISTER", 1,
                                        STAFCommandParser.VALUENOTALLOWED);

            fUnregisterParser.addOption("ID", 1,
                                        STAFCommandParser.VALUEREQUIRED);

            fUnregisterParser.addOptionNeed("UNREGISTER", "ID");

            fListParser.addOption("LIST", 1,
                                  STAFCommandParser.VALUENOTALLOWED);

            fListParser.addOption("MACHINE", 1,
                                  STAFCommandParser.VALUEREQUIRED);

            fListParser.addOption("LONG", 1,
                                  STAFCommandParser.VALUENOTALLOWED);

            fListParser.addOption("SHORT", 1,
                                  STAFCommandParser.VALUENOTALLOWED);

            fListParser.addOptionGroup("LONG SHORT", 0, 1);

            fTriggerParser.addOption("TRIGGER", 1,
                                     STAFCommandParser.VALUENOTALLOWED);

            fTriggerParser.addOption("ID", 1,
                                     STAFCommandParser.VALUEREQUIRED);

            fTriggerParser.addOptionNeed("TRIGGER", "ID");

            fTriggerParser.addOption("SCRIPT", 0,
                                     STAFCommandParser.VALUEREQUIRED);

            fEnableParser.addOption("ENABLE", 1,
                                    STAFCommandParser.VALUENOTALLOWED);

            fEnableParser.addOption("ID", 1,
                                    STAFCommandParser.VALUEREQUIRED);

            fEnableParser.addOptionNeed("ENABLE", "ID");

            fDisableParser.addOption("DISABLE", 1,
                                     STAFCommandParser.VALUENOTALLOWED);

            fDisableParser.addOption("ID", 1,
                                     STAFCommandParser.VALUEREQUIRED);

            fDisableParser.addOptionNeed("DISABLE", "ID");

            fVersionParser.addOption("VERSION", 1,
                                     STAFCommandParser.VALUENOTALLOWED);

            fVersionParser.addOption("JYTHON", 1,
                                     STAFCommandParser.VALUENOTALLOWED);

            fVersionParser.addOptionGroup("VERSION", 1, 1);
            fVersionParser.addOptionNeed("JYTHON", "VERSION");

            fCronIDMapClass = new STAFMapClassDefinition(
                "STAF/Service/Cron/CronID");
            fCronIDMapClass.addKey("cronID", "ID");
            fCronIDMapClass.addKey("description", "Description");
            fCronIDMapClass.setKeyProperty("description", "display-short-name",
                                           "Desc");
            fCronIDMapClass.addKey("machine", "Machine");
            fCronIDMapClass.addKey("service", "Service");
            fCronIDMapClass.addKey("request", "Request");
            fCronIDMapClass.addKey("minute", "Minute");
            fCronIDMapClass.addKey("hour", "Hour");
            fCronIDMapClass.addKey("dayOfMonth", "Day of Month");
            fCronIDMapClass.setKeyProperty("dayOfMonth", "display-short-name",
                                           "Day");
            fCronIDMapClass.addKey("month", "Month");
            fCronIDMapClass.addKey("dayOfWeek", "Day of Week");
            fCronIDMapClass.setKeyProperty("dayOfWeek", "display-short-name",
                                           "Weekday");
            fCronIDMapClass.addKey("once", "Once");

            fCronIDShortMapClass = new STAFMapClassDefinition(
                "STAF/Service/Cron/CronIDShort");
            fCronIDShortMapClass.addKey("cronID", "ID");
            fCronIDShortMapClass.addKey("description", "Description");
            fCronIDShortMapClass.setKeyProperty("description",
                                                "display-short-name",
                                                "Desc");
            fCronIDShortMapClass.addKey("machine", "Machine");
            fCronIDShortMapClass.addKey("service", "Service");
            fCronIDShortMapClass.addKey("request", "Request");

            fCronIDLongMapClass = new STAFMapClassDefinition(
                "STAF/Service/Cron/CronIDLong");
            fCronIDLongMapClass.addKey("cronID", "ID");
            fCronIDLongMapClass.addKey("description", "Description");
            fCronIDLongMapClass.addKey("machine", "Machine");
            fCronIDLongMapClass.addKey("machineType", "Machine Type");
            fCronIDLongMapClass.addKey("service", "Service");
            fCronIDLongMapClass.addKey("serviceType", "Service Type");
            fCronIDLongMapClass.addKey("request", "Request");
            fCronIDLongMapClass.addKey("requestType","Request Type");
            fCronIDLongMapClass.addKey("prepareScript", "Prepare Script");
            fCronIDLongMapClass.addKey("minute", "Minute");
            fCronIDLongMapClass.addKey("hour", "Hour");
            fCronIDLongMapClass.addKey("dayOfMonth", "Day of Month");
            fCronIDLongMapClass.addKey("month", "Month");
            fCronIDLongMapClass.addKey("dayOfWeek", "Day of Week");
            fCronIDLongMapClass.addKey("once", "Once");
            fCronIDLongMapClass.addKey("state", "State");

            // Construct map class for a TRIGGER request

            fTriggerMapClass = new STAFMapClassDefinition(
                "STAF/Service/Cron/Trigger");
            fTriggerMapClass.addKey("machine", "Machine");
            fTriggerMapClass.addKey("requestNumber", "Request Number");

            // Resolve the file separator variable for the local machine

            res = STAFUtil.resolveInitVar("{STAF/Config/Sep/File}", fHandle);

            if (res.rc != STAFResult.Ok) return res;

            String fileSep = res.result;

            // Store data for the cron service in directory:
            //   <STAF writeLocation>/service/<cron service name (lower-case)>

            fHashtableFileDirectory = info.writeLocation;

            if (!fHashtableFileDirectory.endsWith(fileSep))
            {
                fHashtableFileDirectory += fileSep;
            }

            fHashtableFileDirectory = fHashtableFileDirectory +
                "service" + fileSep + fServiceName.toLowerCase();

            File dir = new File(fHashtableFileDirectory);

            if (!dir.exists())
            {
                dir.mkdirs();
            }

            fHashtableFileName = fHashtableFileDirectory + fileSep +
                "cron.ser";

            loadHashtable();

            /* Register RCs with the HELP service */

            rc = this.registerHelp(info.name);

            if (rc != STAFResult.Ok)
                return new STAFResult(rc, 
                    "Error registering RCs with HELP service.");

            fQueueThread = new Thread(this);
            fQueueThread.start(); // this calls the run() method
        }
        catch (STAFException e)
        {
            return new STAFResult(STAFResult.STAFRegistrationError,
                                  e.toString());
        }

        return new STAFResult(rc);
    }

    private int registerHelp(String name)
    {
        try
        {
            String request = "REGISTER SERVICE " + name + " ERROR " +
                PYTHONERROR + " INFO \"" + PYTHONERRORInfo + "\" DESCRIPTION \""
                + PYTHONERRORDesc + "\"";
            fHandle.submit("LOCAL", "HELP", request);

            request = "REGISTER SERVICE " + name + " ERROR " +
                REQUESTNOTSUBMITTED + " INFO \"" + REQUESTNOTSUBMITTEDInfo +
                "\" DESCRIPTION \"" + REQUESTNOTSUBMITTEDDesc + "\"";
            fHandle.submit("LOCAL", "HELP", request);
        }
        catch(STAFException se)
        {
            return se.rc;
        }

        return STAFResult.Ok;
    }

    private STAFResult handleParms(STAFServiceInterfaceLevel30.InitInfo info)
        throws STAFException
    {
        STAFCommandParseResult parseResult = fParmsParser.parse(info.parms);

        if (parseResult.rc != STAFResult.Ok)
        {
            return new STAFResult(
                STAFResult.InvalidRequestString, parseResult.errorBuffer);
        }

        if (parseResult.optionTimes("OLDVARRESOLUTION") > 0)
        {
            fOldVarResolution = true;
        }

        return new STAFResult(STAFResult.Ok);
    }

    public STAFResult acceptRequest(STAFServiceInterfaceLevel30.RequestInfo info)
    {
        String lowerRequest = info.request.toLowerCase();

        // call the appropriate method to handle the command
        if (lowerRequest.startsWith("register"))
        {
            return handleRegister(info);
        }
        else if (lowerRequest.startsWith("unregister"))
        {
            return handleUnregister(info);
        }
        else if (lowerRequest.startsWith("list"))
        {
            return handleList(info);
        }
        else if (lowerRequest.startsWith("trigger"))
        {
            return handleTrigger(info);
        }
        else if (lowerRequest.startsWith("enable"))
        {
            return handleEnable(info);
        }
        else if (lowerRequest.startsWith("disable"))
        {
            return handleDisable(info);
        }
        else if (lowerRequest.startsWith("help"))
        {
            return handleHelp(info);
        }
        else if (lowerRequest.startsWith("version"))
        {
            return handleVersion(info);
        }
        else
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  "Unknown Cron Request: " +
                                  lowerRequest);
        }
    }

    private STAFResult handleHelp(STAFServiceInterfaceLevel30.RequestInfo info)
    {
        // Verify the requester has at least trust level 1

        STAFResult trustResult = STAFUtil.validateTrust(
            1, fServiceName, "HELP", fLocalMachineName, info);

        if (trustResult.rc != STAFResult.Ok) return trustResult;

        // Return help text

        return new STAFResult(STAFResult.Ok,
         "Cron Service Help" + fLineSep
         + "REGISTER   [DESCRIPTION <description>]"
         + fLineSep +
         "           MACHINE <machine> | PYTHONMACHINE <machine>"
         + fLineSep +
         "           SERVICE <service> | PYTHONSERVICE <machine>"
         + fLineSep +
         "           REQUEST <request> | PYTHONREQUEST <request>"
         + fLineSep + "           [PREPARE <script>]"
         + fLineSep + "           [MINUTE <minute>] [HOUR <hour>] "
         + fLineSep + "           [DAY <day>] [MONTH <month>] "
         + fLineSep + "           [WEEKDAY <weekday>]"
         + fLineSep + "           [ONCE]"
         + fLineSep + "           [ENABLED | DISABLED]"
         + fLineSep + "UNREGISTER ID <registrationID>"
         + fLineSep + "LIST       [MACHINE <machine>] [LONG | SHORT]"
         + fLineSep + "TRIGGER    ID <registrationID> [SCRIPT <Python code>]..."
         + fLineSep + "ENABLE     ID <registrationID>"
         + fLineSep + "DISABLE    ID <registrationID>"
         + fLineSep + "VERSION    [JYTHON]" + fLineSep + "HELP");
    }

    private STAFResult handleList(STAFServiceInterfaceLevel30.RequestInfo info)
    {
        STAFCommandParseResult parsedRequest = fListParser.parse(info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(
                STAFResult.InvalidRequestString, parsedRequest.errorBuffer);
        }

        STAFMarshallingContext mc = new STAFMarshallingContext();
        List cronIDList = new ArrayList();

        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        String machineParm = "";
        String typeParm = "";

        try
        {
            // Verify the requester has at least trust level 2

            STAFResult trustResult = STAFUtil.validateTrust(
                2, fServiceName, "LIST", fLocalMachineName, info);

            if (trustResult.rc != STAFResult.Ok) return trustResult;

            // Process the list request

            int registrationTableSize = fRegTable.size();

            if (registrationTableSize > 0)
            {
                // Resolve the MACHINE option

                resolvedValue = STAFUtil.resolveRequestVar(
                    parsedRequest.optionValue("machine"),
                    fHandle, info.requestNumber);

                if (resolvedValue.rc != 0) return resolvedValue;

                machineParm = resolvedValue.result;

                // Check if the LONG or SHORT options are specified

                boolean longFormat = false;
                boolean shortFormat = false;

                if (parsedRequest.optionTimes("LONG") > 0)
                {
                    longFormat = true;
                    mc.setMapClassDefinition(fCronIDLongMapClass);
                }
                else if (parsedRequest.optionTimes("SHORT") > 0)
                {
                    shortFormat = true;
                    mc.setMapClassDefinition(fCronIDShortMapClass);
                }
                else
                {
                    mc.setMapClassDefinition(fCronIDMapClass);
                }

                CronData2 theData;
                boolean match;

                int numberOfKeys = fRegTable.size();
                Integer[] ids = new Integer[numberOfKeys];
                Enumeration idKeys = fRegTable.keys();

                for (int i = 0; idKeys.hasMoreElements(); i++)
                {
                    ids[i] = (Integer)idKeys.nextElement();
                }

                Arrays.sort(ids);

                for (int i = 0; i < numberOfKeys; i++)
                {
                    Integer id = ids[i];
                    match = false;
                    theData = (CronData2) fRegTable.get(id);

                    String description = "";
                    String minutes = "";
                    String hours = "";
                    String days = "";
                    String months = "";
                    String weekdays = "";

                    if (machineParm.equals("") |
                        machineParm.equals(theData.fMachine))
                    {
                        description = theData.fDescription;
 
                        Enumeration minuteEnum = theData.fMinute.elements();

                        while (minuteEnum.hasMoreElements())
                        {
                            minutes += (minuteEnum.nextElement()).toString();

                            if (minuteEnum.hasMoreElements())
                                minutes += ",";
                        }

                        Enumeration hourEnum = theData.fHour.elements();

                        while (hourEnum.hasMoreElements())
                        {
                            hours += (hourEnum.nextElement()).toString();

                            if (hourEnum.hasMoreElements())
                                hours += ",";
                        }

                        Enumeration dayEnum = theData.fDay.elements();

                        while (dayEnum.hasMoreElements())
                        {
                            days += (dayEnum.nextElement()).toString();

                            if (dayEnum.hasMoreElements())
                                days += ",";
                        }

                        Enumeration monthEnum = theData.fMonth.elements();

                        while (monthEnum.hasMoreElements())
                        {
                            months += (monthEnum.nextElement()).toString();

                            if (monthEnum.hasMoreElements())
                                months += ",";
                        }

                        Enumeration weekdayEnum = theData.fWeekday.elements();

                        while (weekdayEnum.hasMoreElements())
                        {
                            weekdays += (weekdayEnum.nextElement()).toString();

                            if (weekdayEnum.hasMoreElements())
                                weekdays += ",";
                        }

                        // Create a map representing the matching registration
                        // and add it to the list of registrations

                        Map cronIDMap;

                        if (longFormat)
                            cronIDMap = fCronIDLongMapClass.createInstance();
                        else if (shortFormat)
                            cronIDMap = fCronIDShortMapClass.createInstance();
                        else
                            cronIDMap = fCronIDMapClass.createInstance();

                        cronIDMap.put("cronID", id.toString());

                        if (!theData.fDescription.equals(""))
                        {
                            cronIDMap.put("description", theData.fDescription);
                        }

                        cronIDMap.put("machine", theData.fMachine);
                        cronIDMap.put("service", theData.fService);
                        cronIDMap.put("request", STAFUtil.maskPrivateData(
                            theData.fRequest));
                        
                        if ((minutes != null) && (!minutes.equals(""))
                            && !shortFormat)
                        {
                            cronIDMap.put("minute", minutes);
                        }
                        if ((hours != null) && (!hours.equals(""))
                            && !shortFormat)
                        {
                            cronIDMap.put("hour", hours);
                        }
                        if ((days != null) && (!days.equals(""))
                            && !shortFormat)
                        {
                            cronIDMap.put("dayOfMonth", days);
                        }
                        if ((months != null) && (!months.equals(""))
                            && !shortFormat)
                        {
                            cronIDMap.put("month", months);
                        }
                        if ((weekdays != null) && (!weekdays.equals(""))
                            && !shortFormat)
                        {
                            cronIDMap.put("dayOfWeek", weekdays);
                        }

                        if (!shortFormat)
                        {
                            cronIDMap.put("once", 
                                new Boolean(theData.fOnce).toString());
                        }

                        if (longFormat)
                        {
                            if (theData.fPythonMachine)
                                cronIDMap.put("machineType", "Python");
                            else
                                cronIDMap.put("machineType", "Literal");

                            if (theData.fPythonService)
                                cronIDMap.put("serviceType", "Python");
                            else
                                cronIDMap.put("serviceType", "Literal");

                            if (theData.fPythonRequest)
                                cronIDMap.put("requestType", "Python");
                            else
                                cronIDMap.put("requestType", "Literal");

                            if ((theData.fPrepare != null) &&
                                (!theData.fPrepare.equals("")))
                            {
                                cronIDMap.put("prepareScript",
                                              STAFUtil.maskPrivateData(
                                                  theData.fPrepare));
                            }

                            cronIDMap.put("state", theData.fState);
                        }

                        cronIDList.add(cronIDMap);
                    }
                }
            }
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }
            return new STAFResult(STAFResult.JavaError,
                                  "Internal Java error.");
        }

        mc.setRootObject(cronIDList);

        return new STAFResult(STAFResult.Ok, mc.marshall());
    }

    private STAFResult handleTrigger(STAFServiceInterfaceLevel30.RequestInfo
                                      info)
    {
        STAFResult resolvedValue = null;
        Integer id;

        STAFCommandParseResult parsedRequest =
            fTriggerParser.parse(info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(
                STAFResult.InvalidRequestString, parsedRequest.errorBuffer);
        }

        try
        {
            // Verify the requester has at least trust level 5

            STAFResult trustResult = STAFUtil.validateTrust(
                5, fServiceName, "TRIGGER", fLocalMachineName, info);

            if (trustResult.rc != STAFResult.Ok) return trustResult;

            // Resolve the ID option

            resolvedValue = STAFUtil.resolveRequestVarAndCheckInt(
                "id", parsedRequest.optionValue("id"),
                fHandle, info.requestNumber);

            if (resolvedValue.rc != 0) return resolvedValue;
            
            id = new Integer(resolvedValue.result);
            
            // Check if registration table contains the ID

            if (!fRegTable.containsKey(id))
            {
                return new STAFResult(
                    STAFResult.DoesNotExist, "ID " + id + " not found");
            }
            else
            {
                // Create a marshalling context and set any map classes (if any)

                STAFMarshallingContext mc = new STAFMarshallingContext();
                mc.setMapClassDefinition(fTriggerMapClass);

                // Create an empty result map to contain the result

                Map resultMap = fTriggerMapClass.createInstance();

                // Handle any SCRIPT options

                int scriptCount = parsedRequest.optionTimes("script");
                LinkedList scripts = new LinkedList();

                if (scriptCount > 0)
                {
                    for (int i = 1; i <= scriptCount; i++)
                    {
                        resolvedValue = STAFUtil.resolveRequestVar(
                            parsedRequest.optionValue("script", i),
                            fHandle, info.requestNumber);

                        scripts.add(resolvedValue.result);
                    }
                }

                String triggerMsg = "[ID=" + id + "] [" + info.endpoint +
                    ", " + info.handleName + "] Triggering a STAF command.\n" +
                    info.request;

                log("info", triggerMsg);

                CronData2 theData = (CronData2)fRegTable.get(id);

                // create new Python Interpreter for each Cron
                fPythonInterpreter = new PythonInterpreter();

                String regInfoMsg = "  ID: " + id.toString() +
                    "\n  machine: " + theData.fMachine +
                    "\n  service: " + theData.fService +
                    "\n  request: " + theData.fRequest +
                    "\n  prepare: " + theData.fPrepare;

                // First execute any SCRIPT options

                for (int j = 0; j < scripts.size(); j++)
                {
                    String script = (String)scripts.get(j);

                    try
                    {
                        pyExec(script);
                    }
                    catch (PyException e)
                    {
                        String pythonError = "Python error in the " +
                            "SCRIPT value.\n\nSCRIPT:\n" + script +
                            "\n\nPyException:\n" + e +
                            "\nProcessing registration:\n" + regInfoMsg;

                        String msg = "[ID=" + id + "] " + pythonError;
 
                        log("error", msg);

                        return new STAFResult(PYTHONERROR, pythonError);
                    }
                }

                pySetVar("STAFCronSubmit", "true");

                String prepare = theData.fPrepare;

                if (prepare != null && !prepare.equals(""))
                {
                    try
                    {
                        pyExec(prepare);
                    }
                    catch (PyException ex)
                    {
                        String pythonError = "Python error in the " +
                            "PREPARE value.\n\nPREPARE:\n" + prepare +
                            "\n\nPyException:\n" + ex +
                            "\nProcessing registration:\n" + regInfoMsg;

                        String msg = "[ID=" + id + "] " + pythonError;
                        
                        log("error", msg);

                        return new STAFResult(PYTHONERROR, pythonError);
                    }

                    Object submit = pyGetVar("STAFCronSubmit");

                    if ((submit == null) ||
                        (!submit.toString().equals("true")))
                    {
                        // Do not submit the request for this registration
                        return new STAFResult(REQUESTNOTSUBMITTED,
                            "STAFCronSubmit=" + submit.toString());
                    }
                }

                boolean pythonMachine = theData.fPythonMachine;
                boolean pythonService = theData.fPythonService;
                boolean pythonRequest = theData.fPythonRequest;

                String machine = "";
                String service = "";
                String request = "";

                if (pythonMachine)
                {
                    try
                    {
                        machine = pyStringEval(theData.fMachine);
                    }
                    catch (PyException ex)
                    {
                        String pythonError = "Python error in the " +
                            "PYTHONMACHINE value.\n\nPYTHONMACHINE: " +
                            theData.fMachine + "\n\nPyException:\n" + ex +
                            "\nProcessing registration:\n" + regInfoMsg;

                        String msg = "[ID=" + id + "] " + pythonError;

                        log("error", msg);

                        return new STAFResult(PYTHONERROR, pythonError);
                    }
                }
                else
                {
                    machine = theData.fMachine;
                }

                if (pythonService)
                {
                    try
                    {
                        service = pyStringEval(theData.fService);
                    }
                    catch (PyException ex)
                    {
                        String pythonError = "Python error in the " +
                            "PYTHONSERVICE value.\n\nPYTHONSERVICE: " +
                            theData.fService + "\n\nPyException:\n" + ex +
                            "\nProcessing registration:\n" + regInfoMsg;

                        String msg = "[ID=" + id + "] " + pythonError;
 
                        log("error", msg);

                        return new STAFResult(PYTHONERROR, pythonError);
                    }
                }
                else
                {
                    service = theData.fService;
                }

                if (pythonRequest)
                {
                    try
                    {
                        request = pyStringEval(theData.fRequest);
                    }
                    catch (PyException ex)
                    {
                        String pythonError = "Python error in the " +
                            "PYTHONREQUEST value.\n\nPYTHONREQUEST: " +
                            theData.fRequest + "\n\nPyException:\n" + ex +
                            "\nProcessing registration:\n" + regInfoMsg;

                        String msg = "[ID=" + id + "] " + pythonError;
                        
                        log("error", msg);

                        return new STAFResult(PYTHONERROR, pythonError);
                    }
                }
                else
                {
                    request = theData.fRequest;
                }

                // Check if a PROCESS START request is being submitted without
                // the WAIT option and without the NOTIFY ONEND option and,
                // if so, add "NOTIFY ONEND Handle <ServiceHandle>" to the
                // request and return the updated request.

                if (service.equalsIgnoreCase("PROCESS"))
                {
                    request = updateProcessStartRequest(request);
                }

                STAFResult submitResult = fHandle.submit2(
                    STAFHandle.ReqQueue, machine, service, request);

                if (submitResult.rc != STAFResult.Ok)
                {
                    // Error submitting the STAF service request

                    String message = "[ID=" + id.toString() + "] " +
                        "Error submitting a STAF command." +
                        "\nTriggered manually." +
                        "\nSTAF command: STAF " + machine + " " +
                        service + " " + request +
                        "\nSubmit RC: " + submitResult.rc +
                        ", Result: " + submitResult.result;
                    
                    log("error", message);

                    return submitResult;
                }

                String requestNumber = submitResult.result;
                
                fSubmittedRequests.put(
                    requestNumber, "machine=" + machine +
                    " service=" + service + " request=" + request);

                // Machine name is used as the key for fRequestsAndIDs map and
                // in a logged messaeg

                String machineName = machine;

                STAFResult varRequest = fHandle.submit2(
                    machine, "VAR", "RESOLVE STRING {STAF/Config/Machine}");

                if (varRequest.rc == STAFResult.Ok)
                    machineName = varRequest.result;

                fRequestsAndIDs.put(machineName + ":" + requestNumber,
                                    id.toString());

                String message = "[ID=" + id.toString() + "] " +
                    "[" + machineName + ":" + requestNumber +
                    "] Submitted a STAF command.\nTriggered manually." +
                    "\nSTAF command: STAF " + machine + " " +
                    service + " " + request;

                log("info", message);

                resultMap.put("machine", machineName);
                resultMap.put("requestNumber", requestNumber);

                // Set the result map as the root object for the marshalling
                // context and return the marshalled result

                mc.setRootObject(resultMap);

                return new STAFResult(STAFResult.Ok, mc.marshall());
            }
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }

            return new STAFResult(STAFResult.JavaError,
                                   "Internal Java error.");
        }
    }

    private STAFResult handleEnable(STAFServiceInterfaceLevel30.RequestInfo
                                     info)
    {
        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        Integer id;

        STAFCommandParseResult parsedRequest =
            fEnableParser.parse(info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(
                STAFResult.InvalidRequestString, parsedRequest.errorBuffer);
        }

        try
        {
            // Verify the requester has at least trust level 4

            STAFResult trustResult = STAFUtil.validateTrust(
                4, fServiceName, "ENABLE", fLocalMachineName, info);

            if (trustResult.rc != STAFResult.Ok) return trustResult;

            // Resolve the ID option

            resolvedValue = STAFUtil.resolveRequestVarAndCheckInt(
                "id", parsedRequest.optionValue("id"),
                fHandle, info.requestNumber);

            if (resolvedValue.rc != 0) return resolvedValue;

            id = new Integer(resolvedValue.result);

            // Check if registration table contains the ID

            if (!fRegTable.containsKey(id))
            {
                return new STAFResult(
                    STAFResult.DoesNotExist, "ID " + id + " not found");
            }

            CronData2 data = (CronData2)fRegTable.get(id);
            data.fState = "Enabled";

            fRegTable.put(id, data);

            saveHashtable();

            String message = "[ID=" + id + "] [" + info.endpoint + ", " +
                    info.handleName + "] Enabled a STAF command.";

            log("info", message);
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }

            return new STAFResult(STAFResult.JavaError,
                                   "Internal Java error.");
        }

        return sResult;
    }

    private STAFResult handleDisable(STAFServiceInterfaceLevel30.RequestInfo
                                      info)
    {
        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        Integer id;

        STAFCommandParseResult parsedRequest =
            fDisableParser.parse(info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(
                STAFResult.InvalidRequestString, parsedRequest.errorBuffer);
        }

        try
        {
            // Verify the requester has at least trust level 4

            STAFResult trustResult = STAFUtil.validateTrust(
                4, fServiceName, "DISABLE", fLocalMachineName, info);

            if (trustResult.rc != STAFResult.Ok) return trustResult;

            // Resolve the ID option

            resolvedValue = STAFUtil.resolveRequestVarAndCheckInt(
                "id", parsedRequest.optionValue("id"),
                fHandle, info.requestNumber);

            if (resolvedValue.rc != 0) return resolvedValue;

            id = new Integer(resolvedValue.result);

            // Check if registration table contains the ID

            if (!fRegTable.containsKey(id))
            {
                return new STAFResult(
                    STAFResult.DoesNotExist, "ID " + id + " not found");
            }

            CronData2 data = (CronData2)fRegTable.get(id);
            data.fState = "Disabled";

            fRegTable.put(id, data);

            saveHashtable();

            String message = "[ID=" + id + "] [" + info.endpoint + ", " +
                    info.handleName + "] Disabled a STAF command.";

            log("info", message);
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }

            return new STAFResult(STAFResult.JavaError,
                                   "Internal Java error.");
        }

        return sResult;
    }

    private STAFResult handleRegister(
                           STAFServiceInterfaceLevel30.RequestInfo info)
    {
        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        String description = "";
        String originMachine = info.endpoint;
        String machine = "";
        boolean pythonMachine = true;
        String service = "";
        boolean pythonService = true;
        String request = "";
        boolean pythonRequest = true;
        String prepare = "";
        boolean once = false;
        String state = "Enabled";

        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        STAFCommandParseResult parsedRequest =
            fRegisterParser.parse(info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parsedRequest.errorBuffer);
        }

        // For each option, if the user has specified a value, set the
        // corresponding local variable to the user-specified value.

        try
        {
            // Verify the requester has at least trust level 5

            STAFResult trustResult = STAFUtil.validateTrust(
                5, fServiceName, "REGISTER", fLocalMachineName, info);

            if (trustResult.rc != STAFResult.Ok) return trustResult;

            // Process the request

            if (parsedRequest.optionTimes("description") > 0)
            {
                resolvedValue = STAFUtil.resolveRequestVar(
                    parsedRequest.optionValue("description"),
                    fHandle, info.requestNumber);
                    
                description = resolvedValue.result;
            }

            if (parsedRequest.optionTimes("machine") > 0)
            {
                machine = parsedRequest.optionValue("machine");

                if (fOldVarResolution)
                {
                    // Resolve the MACHINE option

                    resolvedValue = STAFUtil.resolveRequestVar(
                        machine, fHandle, info.requestNumber);

                    if (resolvedValue.rc != 0) return resolvedValue;

                    machine = resolvedValue.result;
                }

                pythonMachine = false;
            }
            else if (parsedRequest.optionTimes("pythonmachine") > 0)
            {
                machine = parsedRequest.optionValue("pythonmachine");
                pythonMachine = true;
            }

            if (parsedRequest.optionTimes("service") > 0)
            {
                service = parsedRequest.optionValue("service");

                if (fOldVarResolution)
                {
                    // Resolve the SERVICE option

                    resolvedValue = STAFUtil.resolveRequestVar(
                        service, fHandle, info.requestNumber);

                    if (resolvedValue.rc != 0) return resolvedValue;

                    service = resolvedValue.result;
                }

                pythonService = false;
            }
            else if (parsedRequest.optionTimes("pythonservice") > 0)
            {
                service = parsedRequest.optionValue("pythonservice");
                pythonService = true;
            }

            if (parsedRequest.optionTimes("request") > 0)
            {
                request = parsedRequest.optionValue("request");

                if (fOldVarResolution)
                {
                    // Resolve the REQUEST option

                    resolvedValue = STAFUtil.resolveRequestVar(
                        request, fHandle, info.requestNumber);

                    if (resolvedValue.rc != 0) return resolvedValue;

                    request = resolvedValue.result;
                }

                pythonRequest = false;
            }
            else if (parsedRequest.optionTimes("pythonrequest") > 0)
            {
                request = parsedRequest.optionValue("pythonrequest");
                pythonRequest = true;
            }

            if (parsedRequest.optionTimes("prepare") > 0)
            {
                prepare = parsedRequest.optionValue("prepare");
            }

            if (parsedRequest.optionTimes("enabled") > 0)
            {
                state = "Enabled";
            }
            else if (parsedRequest.optionTimes("disabled") > 0)
            {
                state = "Disabled";
            }

            Vector minute = new Vector();
            Vector hour = new Vector();
            Vector day = new Vector();
            Vector month = new Vector();
            Vector weekday = new Vector();

            if (parsedRequest.optionTimes(kMinute) > 0)
            {
                minute = cronParse(parsedRequest.optionValue(kMinute), kMinute);

                if (minute == null)
                {
                    return new STAFResult(
                        STAFResult.InvalidValue,
                        "Invalid value for " + kMinute + ":  " +
                        parsedRequest.optionValue(kMinute));
                }
            }

            if (parsedRequest.optionTimes(kHour) > 0)
            {
                hour = cronParse(parsedRequest.optionValue(kHour), kHour);

                if (hour == null)
                {
                    return new STAFResult(
                        STAFResult.InvalidValue,
                        "Invalid value for " + kHour + ":  " +
                        parsedRequest.optionValue(kHour));
                }
            }

            if (parsedRequest.optionTimes(kDay) > 0)
            {
                day = cronParse(parsedRequest.optionValue(kDay), kDay);

                if (day == null)
                {
                    return new STAFResult(
                        STAFResult.InvalidValue,
                        "Invalid value for " + kDay + ":  " +
                        parsedRequest.optionValue(kDay));
                }
            }

            if (parsedRequest.optionTimes(kMonth) > 0)
            {
                month = cronParse(parsedRequest.optionValue(kMonth), kMonth);

                if (month == null)
                {
                    return new STAFResult(
                        STAFResult.InvalidValue,
                        "Invalid value for " + kMonth + ":  " +
                        parsedRequest.optionValue(kMonth));
                }
            }

            if (parsedRequest.optionTimes(kWeekday) > 0)
            {
                weekday = cronParse(parsedRequest.optionValue(kWeekday),
                                    kWeekday);

                if (weekday == null)
                {
                    return new STAFResult(
                        STAFResult.InvalidValue,
                        "Invalid value for " + kWeekday + ":  " +
                        parsedRequest.optionValue(kWeekday));
                }
            }

            if (parsedRequest.optionTimes("once") > 0)
            {
                once = true;
            }

            if (!(prepare.equals("")) || pythonMachine || pythonService ||
                pythonRequest)
            {
                // validate Python code

                fPythonInterpreter = new PythonInterpreter();
                
                if (!(prepare.equals("")))
                {
                    try
                    {
                        compileForPython(prepare);
                    }
                    catch (PyException ex)
                    {
                        String msg = "Python code compile failed for the " +
                            "PREPARE value:\n" + prepare +
                            "\n\nPyException:\n" + ex;

                        return new STAFResult(PYTHONERROR, msg);
                    }
                }

                if (pythonMachine)
                {
                    try
                    {
                        compileForPython(machine);
                    }
                    catch (PyException ex)
                    {
                        String msg = "Python code compile failed for the " +
                            "MACHINE value:\n" + machine +
                            "\n\nPyException:\n" + ex;

                            return new STAFResult(PYTHONERROR, msg);
                    }
                }

                if (pythonService)
                {
                    try
                    {
                        compileForPython(service);
                    }
                    catch (PyException ex)
                    {
                        String msg = "Python code compile failed for the " +
                            "SERVICE value:\n" + service +
                            "\n\nPyException:\n" + ex;

                            return new STAFResult(PYTHONERROR, msg);
                    }
                }

                if (pythonRequest)
                {
                    try
                    {
                        compileForPython(request);
                    }
                    catch (PyException ex)
                    {
                        String msg = "Python code compile failed for the " +
                            "REQUEST value:\n" + request +
                            "\n\nPyException:\n" + ex;

                            return new STAFResult(PYTHONERROR, msg);
                    }
                }
            }

            Integer index = new Integer(fCronID++);

            fRegTable.put(
                index,
                (Object) new CronData2(description, originMachine, machine,
                                       pythonMachine, service, pythonService,
                                       request, pythonRequest, prepare, minute,
                                       hour, day, month, weekday, once,
                                       state));

            sResult = new STAFResult(STAFResult.Ok, index.toString());
            saveHashtable();

            String message = "[ID=" + index + "] [" + info.endpoint + ", " +
                info.handleName +
                "] Registered a STAF command.\nRegister request: " +
                info.request;

            log("info", message);
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }
            return new STAFResult(STAFResult.JavaError,
                                  "Internal Java error.");
        }

        return sResult;
    }

    private STAFResult handleUnregister(
        STAFServiceInterfaceLevel30.RequestInfo info)
    {
        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        String service;
        String request;
        Integer id;

        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        STAFCommandParseResult parsedRequest =
            fUnregisterParser.parse(info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parsedRequest.errorBuffer);
        }

        // For each option, if the user has specified a value, set the
        // corresponding local variable to the user-specified value.

        try
        {
            // Verify the requester has at least trust level 4

            STAFResult trustResult = STAFUtil.validateTrust(
                4, fServiceName, "UNREGISTER", fLocalMachineName, info);

            if (trustResult.rc != STAFResult.Ok) return trustResult;

            // Process the request

            resolvedValue = STAFUtil.resolveRequestVar(
                parsedRequest.optionValue("id"), fHandle, info.requestNumber);

            if (resolvedValue.rc != 0) return resolvedValue;
            
            id = new Integer(resolvedValue.result);

            if (!fRegTable.containsKey(id))
            {
                return new STAFResult(STAFResult.DoesNotExist,
                                      "ID " + id + " not found");
            }
            else
            {
                CronData2 data = (CronData2)fRegTable.get(id);

                fRegTable.remove(id);

                if (fRegTable.isEmpty())
                {
                    fCronID = 1;
                }

                saveHashtable();

                String message = "[ID=" + id + "] [" + info.endpoint + ", " +
                    info.handleName + "] Unregistered a STAF command.";

                log("info", message);
            }
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }

            return new STAFResult(
                STAFResult.JavaError, "Internal Java error.");
        }

        return sResult;
    }

    private STAFResult handleVersion(
        STAFServiceInterfaceLevel30.RequestInfo info)
    {
        // Verify the requester has at least trust level 1

        STAFResult trustResult = STAFUtil.validateTrust(
            1, fServiceName, "VERSION", fLocalMachineName, info);

        if (trustResult.rc != STAFResult.Ok) return trustResult;

        // Parse the request

        STAFCommandParseResult parsedRequest = fVersionParser.parse(
            info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parsedRequest.errorBuffer);
        }

        if (parsedRequest.optionTimes("JYTHON") == 0)
        {
            // Return the version of the service
            return new STAFResult(STAFResult.Ok, kVersion);
        }
        else
        {
            // Return the version of Python packaged with the service
            return new STAFResult(STAFResult.Ok, fJythonVersion);
        }
    }

    private Vector cronParse(String msg, String type)
    {
        Vector result = new Vector();
        
        if (msg.equalsIgnoreCase("ANY") || msg.equals("*"))
        {
            result.add("ANY");
            return result;
        }
                
        StringTokenizer options = new StringTokenizer(msg, ",");
            
        while (options.hasMoreElements())
        {
            String element = ((String)options.nextElement()).trim();
        
            // Check if a range was specified:  <start>-<end>

            int firstDashIndex = element.indexOf("-");

            if (firstDashIndex == -1)
            {
                // Single value specified.  Verify that it is valid.
                    
                if (!isValidValue(element, type))
                    return null;

                if (type.equals(kWeekday))
                {
                    int numericWeekday = getNumericWeekday(element);

                    if (numericWeekday == -1)
                    {
                        result.add(new
                            String(new Integer(element).toString()));
                    }
                    else
                    {
                        result.add(new String(
                            new Integer(numericWeekday).toString()));
                    }
                }
                else
                {
                    result.add(new String(new Integer(element).toString()));
                }
            }
            else
            {
                // Range was specified

                if (element.indexOf("-", firstDashIndex + 1) > -1)
                {
                    // Invalid.  Can only contain one dash.
                    return null;
                }

                StringTokenizer items = new StringTokenizer(element, "-");

                String start = null;
                String end = null;

                while (items.hasMoreElements())
                {
                    if (start == null)
                        start = ((String)items.nextElement()).trim();
                    else
                        end = ((String)items.nextElement()).trim();
                }

                // Verify that the values specified for the range are valid
                
                if (!isValidValue(start, type))
                    return null;

                if (!isValidValue(end, type))
                    return null;

                // Types other than weekday will always contain numeric values
                // A Weekday type can contain a numeric value or a string value

                if (!type.equals(kWeekday))
                {
                    // Type is kMonth, kDay, kHour, or kMinute

                    // Iterate through the numeric range

                    int startNum = new Integer(start).intValue();
                    int endNum = new Integer(end).intValue();

                    for (int i = startNum; i <= endNum; i++)
                    {
                        result.add(new String(new Integer(i).toString()));
                    }
                }
                else
                {
                    // Type is kWeekday

                    // Check if using string or numeric values for weekday.

                    boolean startIsNumeric = true;
                    boolean endIsNumeric = true;
                        
                    if (equalsWeekday(start)) startIsNumeric = false;
                    if (equalsWeekday(end)) endIsNumeric = false;
                        
                    int startNum = 0;
                    int endNum = 0;

                    // Get the corresponding numeric value for weekday if
                    // specified as a string, e.g. 0 for "Sunday".

                    if (startIsNumeric)
                        startNum = new Integer(start).intValue();
                    else
                        startNum = getNumericWeekday(start);

                    if (endIsNumeric)
                        endNum = new Integer(end).intValue();
                    else
                        endNum = getNumericWeekday(end);
                        
                    // Iterate throught the range, using the String or numeric
                    // representation based on the start value.

                    for (int i = startNum; i <= endNum; i++)
                    {
                            result.add(new String(new Integer(i).toString()));
                    }
                }
            }
        }

        return result;
    }
    
    boolean isValidValue(String value, String type)
    {
        if (value == null)
            return false;

        int valueNum = 0;

        if (type.equals(kWeekday))
        {
            if (equalsWeekday(value))
                return true;
        }
            
        try
        {
            valueNum = new Integer(value).intValue();
        }
        catch (NumberFormatException ex)
        {
            return false;
        }
        
        if (type.equals(kMinute))
        {
            if (valueNum >= 0 && valueNum <= 59)
                return true;
            else
                return false;
        }
        else if (type.equals(kHour))
        {
            if (valueNum >= 0 && valueNum <= 23)
                return true;
            else
                return false;
        }
        else if (type.equals(kDay))
        {
            if (valueNum >= 1 && valueNum <= 31)
                return true;
            else
                return false;
        }
        else if (type.equals(kMonth))
        {
            if (valueNum >= 1 && valueNum <= 12)
                return true;
            else
                return false;
        }
        else if (type.equals(kWeekday))
        {
            if (valueNum >= 0 && valueNum <= 6)
                return true;
            else
                return false;
        }
        else
        {
            return false;
        }
    }

    private boolean equalsWeekday(String msg)
    {
         if ((msg.equalsIgnoreCase("Sunday")) ||
             (msg.equalsIgnoreCase("Monday")) ||
             (msg.equalsIgnoreCase("Tuesday")) ||
             (msg.equalsIgnoreCase("Wednesday")) ||
             (msg.equalsIgnoreCase("Thursday")) ||
             (msg.equalsIgnoreCase("Friday")) ||
             (msg.equalsIgnoreCase("Saturday")))
         {
            return true;
         }
         else
         {
            return false;
         }
    }

    private int getNumericWeekday(String weekday)
    {
         if (weekday.equalsIgnoreCase("Sunday"))
             return 0;
         else if (weekday.equalsIgnoreCase("Monday"))
             return 1;
         else if (weekday.equalsIgnoreCase("Tuesday"))
             return 2;
         else if (weekday.equalsIgnoreCase("Wednesday"))
             return 3;
         else if (weekday.equalsIgnoreCase("Thursday"))
             return 4;
         else if (weekday.equalsIgnoreCase("Friday"))
             return 5;
         else if (weekday.equalsIgnoreCase("Saturday"))
             return 6;
         else
             return -1;
    }

    private String getStringWeekday(int weekday)
    {
         if (weekday == 0)
             return "Sunday";
         else if (weekday == 1)
             return "Monday";
         else if (weekday == 2)
             return "Tuesday";
         else if (weekday == 3)
             return "Wednesday";
         else if (weekday == 4)
             return "Thursday";
         else if (weekday == 5)
             return "Friday";
         else if (weekday == 6)
             return "Saturday";
         else
             return "";
    }

    private void log(String level, String message)
    {
        fHandle.submit2(
            "local", "LOG", "LOG MACHINE LOGNAME " + fServiceName +
            " LEVEL " + level + " MESSAGE " + STAFUtil.wrapData(message));
    }

    private String updateProcessStartRequest(String request)
    {
        // Check if a PROCESS START request is being submitted without
        // the WAIT option and without the NOTIFY ONEND option.  If so,
        // add "NOTIFY ONEND HANDLE <handle>" to the PROCESS START request
        // so that the service will be notified when the process completes
        // and so the service won't have to submit a PROCESS REGISTER ONEND
        // request later (which may be too late if the process completes
        // very quickly).
        
        String lowerRequest = request.toLowerCase();
        String startString = "start ";
         
        if ((lowerRequest.indexOf(startString) == 0) &&
            (lowerRequest.indexOf(" wait") == -1) &&
            (lowerRequest.indexOf(" notify onend") == -1))
        {
            // Insert "NOTIFY ONEND HANDLE <handle>" after "START " and before
            // the rest of the request to minimize the chance of messing up
            // the request syntax if the last option value isn't "wrapped".
 
            int startLength = startString.length();

            String newRequest = request.substring(
                0, startLength) + "NOTIFY ONEND HANDLE " +
                fHandle.getHandle() + " " + request.substring(startLength);

            return newRequest;
        }
        else
        {
            return request;
        }
    }

    private void handleRequestCompleteMsg(STAFQueueMessage queueMessage)
    {
        // A STAF/RequestComplete message is a map containing keys:
        //   requestNumber, rc, result
        
        Map messageMap = (Map)queueMessage.message;
        
        String reqNum = (String)messageMap.get("requestNumber");
        String rc = (String)messageMap.get("rc");
        Object result = (Object)messageMap.get("result");
         
        // Machine name is used in the key for the fRequestsAndIDs
        // and the fProcessHandlesAndRequests maps and in logged messages
        
        String machineName = queueMessage.machine;
        
        STAFResult varRequest = fHandle.submit2(
            queueMessage.machine, "VAR",
            "RESOLVE STRING {STAF/Config/Machine}");
                    
        if (varRequest.rc == STAFResult.Ok)
             machineName = varRequest.result;
       
        // Log a message about the completed request
        
        String level = "";
        
        if (rc.equals("0"))
            level = "Pass";
        else
            level = "Fail";
        
        String id = (String)(fRequestsAndIDs.get(machineName + ":" + reqNum));
        
        String message = "[ID=" + id + "] [" + machineName + ":" + reqNum +
            "] Completed a STAF command. RC=" + rc + ", Result=" +
            result.toString();

        log(level, message);
        
        // Check if a START request was successfully submitted to the PROCESS
        // service without the WAIT option
        
        String request = (String)fSubmittedRequests.get(reqNum);
        String lowerRequest = request.toLowerCase();
        
        if (rc.equals("0") &&
            (lowerRequest.indexOf("service=process") > -1) &&
            (lowerRequest.indexOf("request=start") > -1)  &&
            (lowerRequest.indexOf(" wait") == -1))
        {
            // Submitted an asynchronous PROCESS START request

            String handle = result.toString();
 
            fProcessHandlesAndRequests.put(machineName + ":" + handle,
                                           machineName + ":" + reqNum);
            
            // Check if the the PROCESS START request contains options to
            // notify the service's handle when the process ends
            
            if (lowerRequest.indexOf(" notify onend handle " +
                                     fHandle.getHandle()) == -1)
            {
                // Not already being notified, so need to register to be
                // notified when the process ends
                
                STAFResult res = fHandle.submit2(
                    queueMessage.machine, "PROCESS",
                    "NOTIFY REGISTER ONENDOFHANDLE " + handle);
                
                if (res.rc != 0)
                {
                    if (res.rc == STAFResult.ProcessAlreadyComplete)
                    {
                        message = "[ID=" + id + "] [" + machineName + ":" +
                            reqNum + "] Process with handle " + handle +
                            " completed before the CRON service could ask " +
                            "to be notified when the process completed so " +
                            "no process completion info can be logged. " +
                            "To fix this, change the registration entry by " +
                            "adding the WAIT option to the PROCESS START " +
                            "request, as well as the RETURNSTDOUT and " +
                            "STDERRTOSTDOUT options if you want the " +
                            "process's stdout/stderr returned and logged.";

                        log("error", message);
                        
                        res = fHandle.submit2(
                            queueMessage.machine, "PROCESS",
                            "FREE HANDLE " + handle);
                    }
                    else
                    {
                        message = "[ID=" + id + "] [" + machineName + ":" +
                            reqNum + "] Process completion notification to " +
                            "machine " + queueMessage.machine +
                            " failed for handle " + handle + " RC=" + res.rc +
                            " Result=" + res.result;
                        
                        log("error", message);
                    }
                }
            }
        }
        else
        {
            // Remove this for requests other than asynchronous PROCESS START
            // requests
            
            fRequestsAndIDs.remove(machineName + ":" + reqNum);
        }
        
        // Remove the request from the submitted request map
        
        fSubmittedRequests.remove(reqNum);
    }

    private void handleProcessEndMsg(STAFQueueMessage queueMessage)
    {
        Map messageMap = (Map)queueMessage.message;
        String handle = (String)messageMap.get("handle");
        String rc = (String)messageMap.get("rc");

        // Machine name is used in the key for the fRequestsAndIDs
        // and the fProcessHandlesAndRequests maps and in logged messages

        String machineName = queueMessage.machine;

        STAFResult varRequest = fHandle.submit2(
            queueMessage.machine, "VAR",
            "RESOLVE STRING {STAF/Config/Machine}");

        if (varRequest.rc == STAFResult.Ok)
            machineName = varRequest.result;

        String request = (String)(fProcessHandlesAndRequests.get(
            machineName + ":" + handle));

        String id = (String)(fRequestsAndIDs.get(request));

        String reqNum = request.substring(request.indexOf(":") + 1);

        String message = "[ID=" + id.toString() + "] " +
            "[" + machineName + ":" + reqNum + "] " +
            "Process completed.\nProcess info: " + messageMap.toString();

        String level = "";

        if (rc.equals("0"))
            level = "Pass";
        else
            level = "Fail";

        log(level, message);

        STAFResult freeHandleResult = fHandle.submit2(
            queueMessage.machine, "PROCESS", "FREE HANDLE " + handle);
    }

    public void run()
    {
        Calendar initialTimestamp = Calendar.getInstance();

        int lastMinute = initialTimestamp.get(Calendar.MINUTE);
        int lastHour = initialTimestamp.get(Calendar.HOUR_OF_DAY);
        int lastDay = initialTimestamp.get(Calendar.DAY_OF_MONTH);
          // MONTH is 0-based, so add 1
        int lastMonth = initialTimestamp.get(Calendar.MONTH) + 1;
        int lastWeekday = initialTimestamp.get(Calendar.DAY_OF_WEEK);

        int currentMinute;
        int currentHour;
        int currentDay;
        int currentMonth;
        int currentWeekday;

        boolean minuteChanged = false;
        boolean hourChanged = false;
        boolean dayChanged = false;
        boolean monthChanged = false;
        boolean weekdayChanged = false;

        STAFResult queueGetResult;

        for (;;)
        {
            queueGetResult = fHandle.submit2(
                "local", "QUEUE", "GET WAIT 10000");

            if (queueGetResult.rc == 0)
            {
                // Need a try/catch block so can catch any errors and continue
                // processing messages on the queue

                try
                {
                    STAFQueueMessage queueMessage = new STAFQueueMessage(
                        queueGetResult.result);

                    String queueType = queueMessage.type;
 
                    if (queueType.equalsIgnoreCase(sQueueTypeEnd))
                    {
                        // This type of queued message indicates that the
                        // service is terminating

                        try
                        {
                            fHandle.unRegister();
                        }
                        catch (STAFException ex)
                        {
                            if (DEBUG)
                            {
                                ex.printStackTrace();
                            }
                        }

                        return;
                    }
                    else if (queueType.equalsIgnoreCase("STAF/RequestComplete"))
                    {
                        handleRequestCompleteMsg(queueMessage);
                    }
                    else if (queueType.equalsIgnoreCase("STAF/PROCESS/END"))
                    {
                        handleProcessEndMsg(queueMessage);
                    }
                }
                catch (Exception e)
                {
                    String msg = "Exception handling queued message. " +
                        "Exception:\n" + e + "\nQueued Message:\n" +
                        queueGetResult.result;

                    log("error", msg);
                }
            }

            // Check if the current time matches any registered triggers and,
            // if so, trigger those Cron registrations.

            minuteChanged = false;
            hourChanged = false;
            dayChanged = false;
            monthChanged = false;
            weekdayChanged = false;

            Calendar currentTimestamp = Calendar.getInstance();

            currentMinute = currentTimestamp.get(Calendar.MINUTE);
            currentHour = currentTimestamp.get(Calendar.HOUR_OF_DAY);
            currentDay = currentTimestamp.get(Calendar.DAY_OF_MONTH);
              // MONTH is 0-based, so add 1
            currentMonth = currentTimestamp.get(Calendar.MONTH) + 1;
            currentWeekday = currentTimestamp.get(Calendar.DAY_OF_WEEK);

            // Calender.SUNDAY -> Calendar.SATURDAY are not guaranteed to be
            // static values across all JVMs and platforms, so determine the
            // actual weekday and set it our 0-based syntax

            if (currentWeekday == java.util.Calendar.SUNDAY)
            {
                currentWeekday = 0;
            }
            else if (currentWeekday == java.util.Calendar.MONDAY)
            {
                currentWeekday = 1;
            }
            else if (currentWeekday == java.util.Calendar.TUESDAY)
            {
                currentWeekday = 2;
            }
            else if (currentWeekday == java.util.Calendar.WEDNESDAY)
            {
                currentWeekday = 3;
            }
            else if (currentWeekday == java.util.Calendar.THURSDAY)
            {
                currentWeekday = 4;
            }
            else if (currentWeekday == java.util.Calendar.FRIDAY)
            {
                currentWeekday = 5;
            }
            else if (currentWeekday == java.util.Calendar.SATURDAY)
            {
                currentWeekday = 6;
            }

            if (currentMinute != lastMinute)
            {
                minuteChanged = true;
                lastMinute = currentMinute;
            }
            else
            {
                minuteChanged = false;
            }

            if (currentHour != lastHour)
            {
                hourChanged = true;
                lastHour = currentHour;
            }
            else
            {
                hourChanged = false;
            }

            if (currentDay != lastDay)
            {
                dayChanged = true;
                lastDay = currentDay;
            }
            else
            {
                dayChanged = false;
            }

            if (currentMonth != lastMonth)
            {
                monthChanged = true;
                lastMonth = currentMonth;
            }
            else
            {
                monthChanged = false;
            }

            if (currentWeekday != lastWeekday)
            {
                weekdayChanged = true;
                lastWeekday = currentWeekday;
            }
            else
            {
                weekdayChanged = false;
            }

            if (minuteChanged || hourChanged || dayChanged ||
                monthChanged || weekdayChanged)
            {
                CronData2 theData;

                for (Enumeration e = fRegTable.keys(); e.hasMoreElements();)
                {
                    Integer id = (Integer)e.nextElement();

                    theData = (CronData2) fRegTable.get(id);

                    boolean minuteMatch = (theData.fMinute.contains(
                        String.valueOf(currentMinute)) ||
                        theData.fMinute.contains("ANY"));

                    boolean minuteMatchOrEmpty = theData.fMinute.isEmpty() ||
                        minuteMatch;

                    boolean hourMatch = (theData.fHour.contains(
                        String.valueOf(currentHour)) ||
                        theData.fHour.contains("ANY"));

                    boolean hourMatchOrEmpty = theData.fHour.isEmpty() ||
                        hourMatch;

                    boolean dayMatch = (theData.fDay.contains(
                        String.valueOf(currentDay)) ||
                        theData.fDay.contains("ANY"));

                    boolean dayMatchOrEmpty = theData.fDay.isEmpty() ||
                        dayMatch;

                    boolean monthMatch = (theData.fMonth.contains(
                        String.valueOf(currentMonth)) ||
                        theData.fMonth.contains("ANY"));

                    boolean monthMatchOrEmpty = theData.fMonth.isEmpty() ||
                        monthMatch;

                    boolean weekdayMatch = (theData.fWeekday.contains(
                        String.valueOf(currentWeekday)) ||
                        theData.fWeekday.contains("ANY"));

                    boolean weekdayMatchOrEmpty = theData.fWeekday.isEmpty() ||
                        weekdayMatch;

                    boolean minuteTrigger = minuteChanged && minuteMatch;

                    boolean hourTrigger = hourChanged && hourMatch;

                    boolean dayTrigger = dayChanged && dayMatch;

                    boolean monthTrigger = monthChanged && monthMatch;

                    boolean weekdayTrigger = weekdayChanged && weekdayMatch;

                    // All 5 time units must match or be empty, and there
                    // must be at least one trigger.
                    if ((minuteMatchOrEmpty && hourMatchOrEmpty &&
                         dayMatchOrEmpty && monthMatchOrEmpty &&
                         weekdayMatchOrEmpty) &&
                        (minuteTrigger || hourTrigger || dayTrigger ||
                         monthTrigger || weekdayTrigger))
                    {
                        // create new Python Interpreter for each Cron
                        fPythonInterpreter = new PythonInterpreter();

                        String regInfoMsg = "  ID: " + id.toString() +
                            "\n  machine: " + theData.fMachine +
                            "\n  service: " + theData.fService +
                            "\n  request: " + theData.fRequest +
                            "\n  prepare: " + theData.fPrepare;

                        pySetVar("STAFCronSubmit", "true");

                        String prepare = theData.fPrepare;

                        if (prepare != null && !prepare.equals(""))
                        {
                            try
                            {
                                pyExec(prepare);
                            }
                            catch (PyException ex)
                            {
                                String msg = "[ID=" + id + 
                                    "] Python error in the PREPARE value." +
                                    "\n\nPREPARE:\n" + prepare +
                                    "\n\nPyException:\n" + ex +
                                    "\nProcessing registration:\n" +
                                    regInfoMsg;

                                log("error", msg);
 
                                continue;
                            }

                            Object submit = pyGetVar("STAFCronSubmit");

                            if ((submit == null) ||
                                (!submit.toString().equals("true")))
                            {
                                // Do not submit the request for
                                //this registration
                                continue;
                            }
                        }

                        boolean pythonMachine = theData.fPythonMachine;
                        boolean pythonService = theData.fPythonService;
                        boolean pythonRequest = theData.fPythonRequest;
                        String state = theData.fState;

                        String machine = "";
                        String service = "";
                        String request = "";

                        if (pythonMachine)
                        {
                            try
                            {
                                machine = pyStringEval(theData.fMachine);
                            }
                            catch (PyException ex)
                            {
                                String msg = "[ID=" + id + "] Python error " +
                                    "in the MACHINE value." +
                                    "\n\nMACHINE:\n" + theData.fMachine +
                                    "\n\nPyException:\n" + ex +
                                    "\nProcessing registration:\n" +
                                    regInfoMsg;

                                log("error", msg);
                                
                                continue;
                            }
                        }
                        else
                        {
                            machine = theData.fMachine;
                        }

                        if (pythonService)
                        {
                            try
                            {
                                service = pyStringEval(theData.fService);
                            }
                            catch (PyException ex)
                            {
                                String msg = "[ID=" + id + "] Python error " +
                                    "in the SERVICE value." +
                                    "\n\nSERVICE:\n" + theData.fService +
                                    "\n\nPyException:\n" + ex +
                                    "\nProcessing registration:\n" +
                                    regInfoMsg;

                                log("error", msg);
                                
                                continue;
                            }
                        }
                        else
                        {
                            service = theData.fService;
                        }

                        if (pythonRequest)
                        {
                            try
                            {
                                request = pyStringEval(theData.fRequest);
                            }
                            catch (PyException ex)
                            {
                                String msg = "[ID=" + id + "] Python error " +
                                    "in the REQUEST value." +
                                    "\n\nREQUEST:\n" + theData.fRequest +
                                    "\n\nPyException:\n" + ex +
                                    "\nProcessing registration:\n" +
                                    regInfoMsg;

                                log("error", msg);
                                
                                continue;
                            }
                        }
                        else
                        {
                            request = theData.fRequest;
                        }

                        String triggers = "";

                        if (minuteTrigger)
                        {
                            triggers += "minute[" + currentMinute + "],";
                        }

                        if (hourTrigger || hourMatch)
                        {
                            triggers += "hour[" + currentHour + "],";
                        }

                        if (dayTrigger || dayMatch)
                        {
                            triggers += "day[" + currentDay + "],";
                        }

                        if (monthTrigger || monthMatch)
                        {
                            triggers += "month[" + currentMonth + "],";
                        }

                        if (weekdayTrigger || weekdayMatch)
                        {
                            triggers += "weekday[" + currentWeekday + "],";
                        }

                        if (triggers.endsWith(","))
                        {
                            triggers = triggers.substring(0,
                                triggers.length() - 1);
                        }

                        if (state.equals("Enabled"))
                        {
                            // Check if a PROCESS START request is being
                            // submitted without the WAIT option and without
                            // the NOTIFY ONEND option and, if so, add
                            // "NOTIFY ONEND Handle <ServiceHandle>" to the
                            // request and return the updated request.
                            
                            if (service.equalsIgnoreCase("PROCESS"))
                            {
                                request = updateProcessStartRequest(request);
                            }
                            
                            STAFResult submitResult = fHandle.submit2(
                                STAFHandle.ReqQueue,
                                machine, service, request);

                            if (submitResult.rc == STAFResult.Ok)
                            {
                                String requestNumber = submitResult.result;

                                fSubmittedRequests.put(
                                    requestNumber, "machine=" + machine +
                                    " service=" + service +
                                    " request=" + request);

                                // Machine name is used in the key for the
                                // fRequestsAndIDs map and in logged messages

                                String machineName = machine;

                                STAFResult varRequest = fHandle.submit2(
                                    machine, "VAR",
                                    "RESOLVE STRING {STAF/Config/Machine}");

                                if (varRequest.rc == STAFResult.Ok)
                                    machineName = varRequest.result;

                                fRequestsAndIDs.put(
                                    machineName + ":" + requestNumber,
                                    id.toString());

                                String message = "[ID=" + id.toString() +
                                    "] " + "[" + machineName + ":" +
                                    requestNumber +
                                    "] Submitted a STAF command." +
                                    "\nTriggers: " + triggers +
                                    "\nSTAF command: STAF " + machine + " " +
                                    service + " " + request;

                                log("info", message);
                            }
                            else
                            {
                                // Error submitting the STAF service request

                                String message = "[ID=" + id.toString() +
                                    "] " + "Error submitting a STAF command." +
                                    "\nTriggers: " + triggers +
                                    "\nSTAF command: STAF " + machine + " " +
                                    service + " " + request +
                                    "\nSubmit RC: " + submitResult.rc +
                                    ", Result: " + submitResult.result;
                                
                                log("error", message);
                            }

                            if (theData.fOnce)
                            {
                                // Remove the registation data

                                fRegTable.remove(id);

                                if (fRegTable.isEmpty())
                                {
                                    fCronID = 1;
                                }

                                saveHashtable();

                                String message = "[ID=" + id +
                                    "] [local://local, " + fServiceHandleName +
                                    "] Unregistered a STAF command.";

                                log("info", message);
                            }
                        }
                        else
                        {
                            String message = "[ID=" + id.toString() + "] " +
                                "ID is disabled. STAF command not submitted." +
                                "\nTriggers: " + triggers +
                                "\nSTAF command: STAF " + machine + " " +
                                service + " " + request;

                            log("info", message);
                        }
                    }
                }
            }
        }

    }

    public void loadHashtable()
    {
        try
        {
            File fHashtableFile = new File(fHashtableFileName);
            if (!fHashtableFile.exists())
            {
                fRegTable = new Hashtable();
                return;
            }
            else
            {
                ObjectInputStream ois =
                    new ObjectInputStream(new FileInputStream(fHashtableFile));
                fRegTable = (Hashtable)(ois.readObject());
                ois.close();

                convertOldCronData();

                fCronID = findNextCronID();

            }
        }
        catch (IOException e)
        {
            fRegTable = new Hashtable();
            return;
        }
        catch (ClassNotFoundException e)
        {
            fRegTable = new Hashtable();
            return;
        }
    }

    public void convertOldCronData()
    {
        // In Cron V3.2.0, a new field, fOnce, was added to the Cron
        // registation data.  A new class was created named CronData1 and will
        // be used in V3.2.0+.  When registering the Cron V3.2.0+ service for
        // the first time, if there are any instances of class CronData in the
        // registration Hashtable, they must be converted from class CronData
        // to CronData1.
        // In Cron V3.3.0, a new field, fState, was added to the Cron
        // registation data.  A new class was created named CronData2 and will
        // be used in V3.3.0+.  When registering the Cron V3.3.0+ service for
        // the first time, if there are any instances of class CronData1 in the
        // registration Hashtable, they must be converted from class CronData1
        // to CronData2.

        Hashtable newRegTable = new Hashtable();

        for (Enumeration e = fRegTable.keys(); e.hasMoreElements();)
        {
            Integer id = (Integer) e.nextElement();

            Object cronDataObj = fRegTable.get(id);

            try
            {
                CronData2 data = (CronData2) cronDataObj;

                newRegTable.put(id, (Object) data);
            }
            catch (ClassCastException ex)
            {
                try
                {
                    CronData1 data = (CronData1) cronDataObj;

                    CronData2 newData = new CronData2(data.fDescription,
                                                      data.fOriginMachine,
                                                      data.fMachine,
                                                      data.fPythonMachine,
                                                      data.fService,
                                                      data.fPythonService,
                                                      data.fRequest,
                                                      data.fPythonRequest,
                                                      data.fPrepare,
                                                      data.fMinute,
                                                      data.fHour,
                                                      data.fDay,
                                                      data.fMonth,
                                                      data.fWeekday,
                                                      data.fOnce,
                                                      "Enabled");

                    newRegTable.put(id, (Object) newData);
                }
                catch (ClassCastException ex2)
                {
                    CronData data = (CronData) cronDataObj;

                    CronData2 newData = new CronData2("",
                                                      data.fOriginMachine,
                                                      data.fMachine,
                                                      data.fPythonMachine,
                                                      data.fService,
                                                      data.fPythonService,
                                                      data.fRequest,
                                                      data.fPythonRequest,
                                                      data.fPrepare,
                                                      data.fMinute,
                                                      data.fHour,
                                                      data.fDay,
                                                      data.fMonth,
                                                      data.fWeekday,
                                                      false,
                                                      "Enabled");

                    newRegTable.put(id, (Object) newData);
                }
            }
        }

        fRegTable = newRegTable;
    }

    public void saveHashtable()
    {
        try
        {
            ObjectOutputStream oos =
                new ObjectOutputStream(new
                    FileOutputStream(fHashtableFileName));
            oos.writeObject(((Object)(fRegTable)));
            oos.close();
        }
        catch (IOException ex)
        {
            ex.printStackTrace();
        }
    }

    public int findNextCronID()
    {
        int highestID = 0;
        for (Enumeration e = fRegTable.keys(); e.hasMoreElements();)
        {
            Integer id = (Integer) e.nextElement();
            if (id.intValue() > highestID)
            {
                highestID = id.intValue();
            }
        }

        return highestID + 1;
    }

    public STAFResult term()
    {
        fHandle.submit2(
            STAFHandle.ReqFireAndForget,
            "local", "QUEUE", "QUEUE TYPE " + sQueueTypeEnd +
            " MESSAGE " + STAFUtil.wrapData(""));

        return new STAFResult(STAFResult.Ok);
    }

    public String pyStringEval(String value)
    {
        synchronized(fPythonInterpreter)
        {
            PyObject result = null;

            result = fPythonInterpreter.eval(value);

            if (result == null)
            {
                return null;
            }
            else
            {
                return result.toString();
            }
        }
    }

    public void pySetVar(String varName, Object value)
    {
        synchronized(fPythonInterpreter)
        {
            try
            {
                fPythonInterpreter.set(varName, value);
            }
            catch (PyException e)
            {
                String msg = "PySetVar PyException caught setting " +
                    " var=" + varName + " to value=" + value + e.toString();

                log("error", msg);
            }
        }
    }

    public Object pyGetVar(String varName)
    {
        synchronized(fPythonInterpreter)
        {
            try
            {
                return fPythonInterpreter.get(varName);
            }
            catch (PyException e)
            {
                String msg = "PyGetVar PyException caught getting " +
                    " var=" + varName + " to value=" + varName + e.toString();

                log("error", msg);

                return null;
            }
        }
    }

    public void pyExec(String value)
    {
        synchronized(fPythonInterpreter)
        {
            fPythonInterpreter.exec(value);
        }
    }
    
    public void compileForPython(String pythonCode)
    {
        PyCode code = __builtin__.compile(pythonCode, "<string>", "exec");
    }

}
