/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002, 2004, 2005                                  */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.email;

import com.ibm.staf.*;
import com.ibm.staf.service.*;
import java.io.*;
import java.util.*;
import java.net.*;
import org.apache.commons.codec.binary.Base64;

public class EmailService implements STAFServiceInterfaceLevel30
{
    private STAFHandle fHandle;
    private final String kVersion = "3.1.2";
    private String fServiceName = "";
    private String fLocalMachineName = "";
    private STAFCommandParser fSendParser;
    private STAFCommandParser fListParser;
    private STAFCommandParser fParmsParser;
    private String fLineSep;
    private STAFMapClassDefinition fSettingsMapClass;
    private static Socket fSocket;
    private static BufferedReader fInStream;
    private static PrintStream fOutStream;
    private static String fHostname;
    private Base64 base64 = new Base64();
    private static String fMailserver = "";
    private static int fMailport = 25;
    static final String sSettingsMapClassName = new String(
        "STAF/Service/Email/Settings");

    public static final int IOEXCEPTION = 4001;
    private static final String IOEXCEPTIONInfo = "IO Exception";
    private static final String IOEXCEPTIONDesc =
       "An IO Exception occured while sending the email.";

    public EmailService() {}

    public STAFResult init(STAFServiceInterfaceLevel30.InitInfo info)
    {
        int rc = STAFResult.Ok;

        try
        {
            fHandle = new STAFHandle("STAF/SERVICE/" + info.name);
        }
        catch (STAFException e)
        {
            return new STAFResult(STAFResult.STAFRegistrationError, e.toString());
        }

        fServiceName = info.name;

        // Send parser

        fSendParser = new STAFCommandParser(0, false);

        fSendParser.addOption("SEND", 1,
                              STAFCommandParser.VALUENOTALLOWED);

        fSendParser.addOption("TO", 0,
                              STAFCommandParser.VALUEREQUIRED);

        fSendParser.addOption("FROM", 1,
                              STAFCommandParser.VALUEREQUIRED);

        fSendParser.addOption("SUBJECT", 1,
                              STAFCommandParser.VALUEREQUIRED);

        fSendParser.addOption("MESSAGE", 1,
                              STAFCommandParser.VALUEREQUIRED);

        fSendParser.addOption("FILE", 1,
                              STAFCommandParser.VALUEREQUIRED);

        fSendParser.addOption("MACHINE", 1,
                              STAFCommandParser.VALUEREQUIRED);

        fSendParser.addOption("NOHEADER", 1,
                              STAFCommandParser.VALUENOTALLOWED);

        fSendParser.addOption("TEXTATTACHMENT", 0,
                              STAFCommandParser.VALUEREQUIRED);

        fSendParser.addOption("BINARYATTACHMENT", 0,
                              STAFCommandParser.VALUEREQUIRED);

        fSendParser.addOption("ATTACHMENTMACHINE", 1,
                              STAFCommandParser.VALUEREQUIRED);

        fSendParser.addOptionGroup("MESSAGE FILE", 1, 1);

        fSendParser.addOptionNeed("MACHINE", "FILE");

        fSendParser.addOptionNeed("ATTACHMENTMACHINE", 
                                  "TEXTATTACHMENT BINARYATTACHMENT");

        // List parser

        fListParser = new STAFCommandParser(0, false);

        fListParser.addOption("LIST", 1, STAFCommandParser.VALUENOTALLOWED);
        fListParser.addOption("SETTINGS", 1, STAFCommandParser.VALUENOTALLOWED);
        fListParser.addOptionNeed("SETTINGS", "LIST");

        // Resolve the line separator variable for the local machine

        STAFResult res = new STAFResult();
         
        res = STAFUtil.resolveInitVar("{STAF/Config/Sep/Line}", fHandle);

        if (res.rc != STAFResult.Ok) return res;
        
        fLineSep = res.result;
         
        // Resolve the machine name variable for the local machine

        res = STAFUtil.resolveInitVar("{STAF/Config/Machine}", fHandle);

        if (res.rc != STAFResult.Ok) return res;
        
        fLocalMachineName = res.result;

        // Parse the parameters specified when registering the service

        if (info.parms == null)
        {
            return new STAFResult(
                STAFResult.ServiceConfigurationError,
                "MAILSERVER is a required parameter for the STAFEmail service");
        }
        else
        {
            fParmsParser = new STAFCommandParser(0, false);

            fParmsParser.addOption("MAILSERVER", 1,
                STAFCommandParser.VALUEREQUIRED);

            fParmsParser.addOption("PORT", 1,
                STAFCommandParser.VALUEREQUIRED);

            // DEBUG is deprecated
            fParmsParser.addOption("DEBUG", 1,
                STAFCommandParser.VALUENOTALLOWED);

            res = handleParms(info);

            if (res.rc != STAFResult.Ok)
            {
                return new STAFResult(
                    STAFResult.ServiceConfigurationError,
                    "Error validating parameters: RC=" + res.rc +
                    ", Result=" + res.result);
            }

            String message = fServiceName + " service initialized, using " +
                "mailserver " + fMailserver + " port " + fMailport;
            fHandle.submit2("local", "LOG", "LOG MACHINE LOGNAME " +
                        fServiceName + " LEVEL info MESSAGE " +
                        STAFUtil.wrapData(message));
        }

        /* Register RCs with the HELP service */

        rc = this.registerHelp(info.name);

        if (rc != STAFResult.Ok)
            return new STAFResult(rc, 
                                   "Error registering RCs with HELP service.");

        // Construct map-class for list settings information

        fSettingsMapClass = new STAFMapClassDefinition(
            sSettingsMapClassName);

        fSettingsMapClass.addKey("mailServer", "Mail Server");
        fSettingsMapClass.addKey("port", "Port");

        return new STAFResult(rc);
    }
    
    private int registerHelp(String name)
    {
        try
        {
            String request = "REGISTER SERVICE " + name + " ERROR " +
                IOEXCEPTION + " INFO \"" + IOEXCEPTIONInfo + "\" DESCRIPTION \""
                + IOEXCEPTIONDesc + "\"";
            fHandle.submit("LOCAL", "HELP", request);
        }
        catch(STAFException se)
        {
            return se.rc;
        }

        return STAFResult.Ok;
    }

    private STAFResult handleParms(STAFServiceInterfaceLevel30.InitInfo info)
    {
        STAFCommandParseResult parseResult= fParmsParser.parse(info.parms);

        if (parseResult.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parseResult.errorBuffer);
        }

        if (parseResult.optionTimes("MAILSERVER") == 0)
        {
            String errorMsg = "MAILSERVER is a required parameter for the " +
                "STAFEmail service";
            return new STAFResult(
                STAFResult.ServiceConfigurationError, errorMsg);
        }
        else
        {
            fMailserver = parseResult.optionValue("MAILSERVER");
        }

        // DEBUG parm is deprecated

        if (parseResult.optionTimes("PORT") > 0)
        {
            fMailport =
                (new Integer(parseResult.optionValue("PORT"))).intValue();
        }

        return new STAFResult(STAFResult.Ok);
    }

    public STAFResult acceptRequest(STAFServiceInterfaceLevel30.RequestInfo info)
    {
        String lowerRequest = info.request.toLowerCase();

        // call the appropriate method to handle the command
        if (lowerRequest.startsWith("send"))
        {
            return handleSend(info);
        }
        else if (lowerRequest.startsWith("list"))
        {
            return handleList(info);
        }
        else if (lowerRequest.startsWith("help"))
        {
            return handleHelp(info);
        }
        else if (lowerRequest.startsWith("version"))
        {
            return handleVersion(info);
        }
        else
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  "Unknown Email Request: " +
                                  lowerRequest);
        }
    }

    private STAFResult handleHelp(STAFServiceInterfaceLevel30.RequestInfo info)
    {
        // Verify the requester has at least trust level 1

        STAFResult trustResult = STAFUtil.validateTrust(
            1, fServiceName, "HELP", fLocalMachineName, info);

        if (trustResult.rc != STAFResult.Ok) return trustResult;

        return new STAFResult(STAFResult.Ok,
         "EMAIL Service Help"
         + fLineSep + fLineSep + "SEND TO <Address>... [FROM <user@company.com>]"
         + fLineSep + "     < MESSAGE <Message> | FILE <File> [MACHINE <Machine>] >"
         + fLineSep + "     [SUBJECT <Subject>] [NOHEADER] [TEXTATTACHMENT <file>]..."
         + fLineSep + "     [BINARYATTACHMENT <file>]... [ATTACHMENTMACHINE <machine>]"
         + fLineSep + fLineSep + "LIST SETTINGS"
         + fLineSep + fLineSep + "VERSION"
         + fLineSep + fLineSep + "HELP");
    }

    private STAFResult handleVersion(STAFServiceInterfaceLevel30.RequestInfo info)
    {
        // Verify the requester has at least trust level 1

        STAFResult trustResult = STAFUtil.validateTrust(
            1, fServiceName, "VERSION", fLocalMachineName, info);

        if (trustResult.rc != STAFResult.Ok) return trustResult;

        return new STAFResult(STAFResult.Ok, kVersion);
    }

    private STAFResult handleList(STAFServiceInterfaceLevel30.RequestInfo info)
    {
        // Verify the requester has at least trust level 2

        STAFResult trustResult = STAFUtil.validateTrust(
            2, fServiceName, "LIST", fLocalMachineName, info);

        if (trustResult.rc != STAFResult.Ok) return trustResult;

        // Parse the result

        STAFCommandParseResult parsedRequest = fListParser.parse(info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parsedRequest.errorBuffer);
        }
          
        // LIST SETTINGS

        // Create the marshalling context

        STAFMarshallingContext mc = new STAFMarshallingContext();

        mc.setMapClassDefinition(fSettingsMapClass);

        Map outputMap = fSettingsMapClass.createInstance();
        outputMap.put("mailServer", fMailserver);
        outputMap.put("port", new Integer(fMailport));

        mc.setRootObject(outputMap);

        return new STAFResult(STAFResult.Ok, mc.marshall());
    }

    private synchronized STAFResult handleSend(
                                   STAFServiceInterfaceLevel30.RequestInfo info)
    {
        // Verify the requester has at least trust level 3

        STAFResult trustResult = STAFUtil.validateTrust(
            3, fServiceName, "SEND", fLocalMachineName, info);

        if (trustResult.rc != STAFResult.Ok) return trustResult;

        STAFResult result = new STAFResult(STAFResult.Ok, "");
        String resultString = "";
        Vector toList = new Vector();
        String messageOption = "";
        String subjectOption = "";
        STAFResult resolvedValue = null;
        Vector textAttachments = new Vector();
        Vector binaryAttachments = new Vector();
        String attachmentMachine = info.endpoint;
        boolean includeHeader = true;

        try
        {
            STAFCommandParseResult parsedRequest = fSendParser.parse(
                info.request);

            if (parsedRequest.rc != STAFResult.Ok)
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                                      parsedRequest.errorBuffer);
            }
            
            int toTimes = parsedRequest.optionTimes("to");

            for (int i = 0; i < toTimes; i++)
            {
                resolvedValue = STAFUtil.resolveRequestVar(
                    parsedRequest.optionValue("to", i + 1),
                    fHandle, info.requestNumber);

                if (resolvedValue.rc != 0) return resolvedValue;

                toList.add(resolvedValue.result);
            }

            if (parsedRequest.optionTimes("subject") > 0)
            {
                resolvedValue = STAFUtil.resolveRequestVar(
                    parsedRequest.optionValue("subject"),
                    fHandle, info.requestNumber);

                if (resolvedValue.rc != 0) return resolvedValue;

                subjectOption = resolvedValue.result;
            }

            if (parsedRequest.optionTimes("noheader") > 0)
            {
                includeHeader = false;
            }

            if (parsedRequest.optionTimes("message") == 1)
            {
                resolvedValue = STAFUtil.resolveRequestVar(
                    parsedRequest.optionValue("message"),
                    fHandle, info.requestNumber);

                if (resolvedValue.rc != 0) return resolvedValue;

                messageOption = resolvedValue.result;
            }
            else if (parsedRequest.optionTimes("file") == 1)
            {
                String fileMachine = info.endpoint;

                if (parsedRequest.optionTimes("machine") == 1)
                {
                    resolvedValue = STAFUtil.resolveRequestVar(
                        parsedRequest.optionValue("machine"),
                        fHandle, info.requestNumber);

                    if (resolvedValue.rc != 0) return resolvedValue;

                    fileMachine = resolvedValue.result;;
                }

                STAFResult res = fHandle.submit2(
                    fileMachine, "FS", "GET FILE " +
                    parsedRequest.optionValue("file"));

                if (res.rc == 0)
                {
                    messageOption = res.result;
                }
                else
                {
                    messageOption = "Error getting email body from machine " +
                    fileMachine + ": RC=" + res.rc + " Result= " + res.result;
                }
            }

            String reply;

            try
            {
                try
                {
                    InetAddress ipaddress = InetAddress.getLocalHost();

                    if ((fHostname = ipaddress.getHostName()) == null)
                    {
                        fHostname = "localhost";
                    }
                    else
                    {
                        ipaddress = InetAddress.getByName(
                            ipaddress.getHostAddress());

                        fHostname = ipaddress.getHostName();
                    }
                }
                catch (UnknownHostException uhe)
                {
                      fHostname = "localhost";
                }

                String from = "STAFEmailService@" + fHostname;

                if (parsedRequest.optionTimes("from") > 0)
                {
                    resolvedValue = STAFUtil.resolveRequestVar(
                        parsedRequest.optionValue("from"),
                        fHandle, info.requestNumber);

                        if (resolvedValue.rc != 0) return resolvedValue;

                        from = resolvedValue.result;
                }

                int index = messageOption.indexOf("\\n");
                String tmpStr = "";
                String CR = "";

                while (index >= 0)
                {
                    tmpStr = tmpStr + CR +
                        messageOption.substring(0, index);
                    CR = "\n";
                    messageOption = messageOption.substring(index+2);
                    index = messageOption.indexOf("\\n");
                }

                tmpStr = tmpStr + CR + messageOption;
                messageOption = tmpStr;

                if (parsedRequest.optionTimes("attachmentmachine") > 0)
                {
                    resolvedValue = STAFUtil.resolveRequestVar(
                        parsedRequest.optionValue("attachmentmachine"),
                        fHandle, info.requestNumber);

                    if (resolvedValue.rc != 0) return resolvedValue;

                    attachmentMachine = resolvedValue.result;
                }

                if (parsedRequest.optionTimes("textattachment") > 0)
                {
                    int textAttachmentCount = 
                        parsedRequest.optionTimes("textattachment");
                        
                    for (int i = 0; i < textAttachmentCount; i++)
                    {
                        resolvedValue = STAFUtil.resolveRequestVar(
                            parsedRequest.optionValue("textattachment", i + 1),
                            fHandle, info.requestNumber);

                        if (resolvedValue.rc != 0) return resolvedValue;
                        
                        textAttachments.add(resolvedValue.result);
                    }
                }

                if (parsedRequest.optionTimes("binaryattachment") > 0)
                {
                    int binaryAttachmentCount = 
                        parsedRequest.optionTimes("binaryattachment");
                        
                    for (int i = 0; i < binaryAttachmentCount; i++)
                    {
                        resolvedValue = STAFUtil.resolveRequestVar(
                            parsedRequest.optionValue("binaryattachment", 
                            i + 1), fHandle, info.requestNumber);

                        if (resolvedValue.rc != 0) return resolvedValue;
                        
                        binaryAttachments.add(resolvedValue.result);
                    }
                }

                String message = "SEND TO " + toList.toString() + " FROM " +
                                 from + " SUBJECT " + subjectOption;

                if (textAttachments.size() > 0)
                {
                    message += " TEXTATTACHMENTS " + textAttachments.toString();
                }

                if (binaryAttachments.size() > 0)
                {
                    message += " BINARYATTACHMENTS " + 
                                binaryAttachments.toString();
                }

                fHandle.submit2(
                    "local", "LOG", "LOG MACHINE LOGNAME " + fServiceName +
                    " LEVEL info MESSAGE " + STAFUtil.wrapData(message));

                for (int j = 0; j < toList.size(); j++)
                {
                    String level = "";
                    String to = (String)toList.elementAt(j);

                    fSocket = new Socket(fMailserver, fMailport);

                    fInStream = new BufferedReader(
                        new InputStreamReader(fSocket.getInputStream()));

                    fOutStream = new PrintStream(fSocket.getOutputStream());

                    reply = fInStream.readLine();
                    level = (reply.startsWith("4") || reply.startsWith("5")) ?
                        "error" : "info";
                    fHandle.submit2("local", "LOG", "LOG MACHINE LOGNAME " +
                        fServiceName + " LEVEL " + level + "  MESSAGE " +
                        STAFUtil.wrapData(reply));

                    if (level.equals("error"))
                        return new STAFResult(
                           new Integer("4" + reply.substring(0, 3)).intValue(),
                           reply);

                    fOutStream.println("HELO " + fHostname);
                    fOutStream.flush();

                    reply = fInStream.readLine();
                    level = (reply.startsWith("4") || reply.startsWith("5")) ?
                        "error" : "info";
                    fHandle.submit2("local", "LOG", "LOG MACHINE LOGNAME " +
                        fServiceName + " LEVEL " + level + "  MESSAGE " +
                        STAFUtil.wrapData(reply));

                    if (level.equals("error"))
                        return new STAFResult(
                           new Integer("4" + reply.substring(0, 3)).intValue(),
                           reply);

                    fOutStream.println("MAIL FROM:  " + from);
                    fOutStream.flush();

                    reply = fInStream.readLine();
                    level = (reply.startsWith("4") || reply.startsWith("5")) ?
                        "error" : "info";
                    fHandle.submit2("local", "LOG", "LOG MACHINE LOGNAME " +
                        fServiceName + " LEVEL " + level + "  MESSAGE " +
                        STAFUtil.wrapData(reply));

                    if (level.equals("error"))
                        return new STAFResult(
                           new Integer("4" + reply.substring(0, 3)).intValue(),
                           reply);

                    fOutStream.println("RCPT TO:  " + to);
                    fOutStream.flush();

                    reply = fInStream.readLine();
                    level = (reply.startsWith("4") || reply.startsWith("5")) ?
                        "error" : "info";
                    fHandle.submit2("local", "LOG", "LOG MACHINE LOGNAME " +
                        fServiceName + " LEVEL " + level + "  MESSAGE " +
                        STAFUtil.wrapData(reply));

                    if (level.equals("error"))
                        return new STAFResult(
                           new Integer("4" + reply.substring(0, 3)).intValue(),
                           reply);

                    fOutStream.println("DATA");
                    fOutStream.flush();

                    // Some external SMTP servers require the following line
                    fInStream.readLine();

                    fOutStream.println("MIME-Version: 1.0");
                    fOutStream.flush();

                    fOutStream.println("To: " + to);
                    fOutStream.flush();

                    fOutStream.println("Subject: " + subjectOption);
                    fOutStream.flush();

                    fOutStream.println("From: " + from);
                    fOutStream.flush();

                    String strBoundary = "DataSeparatorString";

                    fOutStream.println("Content-Type: multipart/mixed; " + 
                        "boundary=\"" + strBoundary + "\"");
                    fOutStream.flush();
                    
                    fOutStream.println("--" + strBoundary);
                    fOutStream.flush();
                    
                    fOutStream.println("Content-Type: text/plain; charset=us-ascii");
                    fOutStream.flush();
                    
                    fOutStream.println("Content-Transfer-Encoding: 7bit");
                    fOutStream.flush();
                    
                    fOutStream.println("");
                    fOutStream.flush();

                    String header =
                        "*****************************************************" +
                        "*************\n"+
                        "* DO NOT RESPOND TO THE SERVICE MACHINE THAT GENERATED " +
                        "THIS NOTE *\n"+
                        "*****************************************************" +
                        "*************\n";

                    if (includeHeader)
                    {
                        fOutStream.println(header);
                        fOutStream.flush();
                    }

                    fOutStream.println(messageOption);
                    fOutStream.flush();

                    fOutStream.println("\n");
                    fOutStream.flush();

                    for (int k = 0; k < textAttachments.size(); k++)
                    {
                        String attachmentFile =
                            (String)textAttachments.elementAt(k);

                        int forwardSlashIndex = attachmentFile.lastIndexOf("/");
                        int backwardSlashIndex =
                            attachmentFile.lastIndexOf("\\");
                        int filenameIndex;

                        if (forwardSlashIndex > backwardSlashIndex)
                            filenameIndex = forwardSlashIndex;
                        else
                            filenameIndex = backwardSlashIndex;

                        String attachmentFileName =
                            attachmentFile.substring(filenameIndex);

                        STAFResult fileResult =
                            fHandle.submit2(attachmentMachine,
                                            "FS", "GET FILE " + attachmentFile +
                                            " text");

                        if (fileResult.rc != STAFResult.Ok)
                            return fileResult;

                        fOutStream.println("--" + strBoundary);
                        fOutStream.flush();

                        fOutStream.println("Content-Type: text/plain; name=\"" +
                            attachmentFileName + "\"");
                        fOutStream.flush();

                        fOutStream.println("Content-Transfer-Encoding: 7bit");
                        fOutStream.flush();

                        fOutStream.println("Content-Disposition: attachment; " +
                            "filename=\"" + attachmentFileName + "\"");
                        fOutStream.flush();

                        fOutStream.println("Content-Description: Form Results");
                        fOutStream.flush();

                        fOutStream.println("");
                        fOutStream.flush();

                        fOutStream.println(fileResult.result);
                        fOutStream.flush();

                        fOutStream.println("--" + strBoundary);
                        fOutStream.flush();

                        fOutStream.println("");
                        fOutStream.flush();
                    }

                    for (int k = 0; k < binaryAttachments.size(); k++)
                    {
                        String attachmentFile =
                            (String)binaryAttachments.elementAt(k);

                        int forwardSlashIndex = attachmentFile.lastIndexOf("/");
                        int backwardSlashIndex =
                            attachmentFile.lastIndexOf("\\");
                        int filenameIndex;

                        if (forwardSlashIndex > backwardSlashIndex)
                            filenameIndex = forwardSlashIndex;
                        else
                            filenameIndex = backwardSlashIndex;

                        String attachmentFileName =
                            attachmentFile.substring(filenameIndex);

                        STAFResult fileResult =
                            fHandle.submit2(attachmentMachine,
                                            "FS", "GET FILE " + attachmentFile +
                                            " binary");

                        if (fileResult.rc != STAFResult.Ok)
                            return fileResult;

                        fOutStream.println("--" + strBoundary);
                        fOutStream.flush();

                        fOutStream.println("Content-Type: " +
                            "application/octet-stream; name=\"" +
                            attachmentFileName + "\"");
                        fOutStream.flush();

                        fOutStream.println("Content-Transfer-Encoding: base64");
                        fOutStream.flush();

                        fOutStream.println("Content-Disposition: attachment; " +
                            "filename=\"" + attachmentFileName + "\"");
                        fOutStream.flush();

                        fOutStream.println("Content-Description: Form Results");
                        fOutStream.flush();

                        fOutStream.println("");
                        fOutStream.flush();

                        String getFileResult = fileResult.result;

                        byte[] fileContent = 
                            new byte[getFileResult.length() / 2];

                        fileContent[0] = (byte)Integer.parseInt(
                            getFileResult.substring(0, 2), 16);

                        for (int i = 1; i < fileContent.length; i++)
                        {
                            fileContent[i] = (byte)
                                Integer.parseInt(getFileResult.substring(
                                                 i*2, i*2 + 2), 16);
                        }

                        byte[] base64FileContent = 
                            base64.encodeBase64Chunked(fileContent);

                        fOutStream.write(base64FileContent, 0,
                                         base64FileContent.length);
                        fOutStream.flush();

                        fOutStream.println("--" + strBoundary);
                        fOutStream.flush();

                        fOutStream.println("");
                        fOutStream.flush();
                    }

                    fOutStream.println("");
                    fOutStream.flush();

                    fOutStream.println();
                    fOutStream.println(".");
                    fOutStream.println();
                    fOutStream.flush();

                    fOutStream.println("QUIT");
                    fOutStream.flush();

                    reply = fInStream.readLine();
                    level = (reply.startsWith("4") || reply.startsWith("5")) ?
                        "error" : "info";
                    fHandle.submit2("local", "LOG", "LOG MACHINE LOGNAME " +
                        fServiceName + " LEVEL " + level + "  MESSAGE " +
                        STAFUtil.wrapData(reply));

                    if (level.equals("error"))
                        return new STAFResult(
                           new Integer("4" + reply.substring(0, 3)).intValue(),
                           reply);

                    fSocket.close();
                }
            }
            catch ( IOException ioe )
            {
                String error = "Socket IO Error: " + ioe +
                    " Mailserver: " + fMailserver + " Port: " + fMailport;

                fHandle.submit2("local", "LOG", "LOG MACHINE LOGNAME " +
                        fServiceName + " LEVEL error MESSAGE " +
                        STAFUtil.wrapData(error));

                return new STAFResult(IOEXCEPTION, error);
            }
        }
        catch (Exception e)
        {
            return new STAFResult(STAFResult.JavaError,
                                  "Internal Java error.");
        }

        return new STAFResult(STAFResult.Ok, resultString);
    }

    public STAFResult term()
    {
        // Un-register the service handle

        try
        {
            fHandle.unRegister();
        }
        catch (STAFException ex)
        {
            return new STAFResult(
                STAFResult.STAFRegistrationError, ex.toString());
        }

        return new STAFResult(STAFResult.Ok);
    }
}
