/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002, 2007                                        */
/*                                                                           */
/* This software is licensed under the Eclipse Public License (EPL) V1.0.    */
/*****************************************************************************/

package com.ibm.staf.service.stax;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.ibm.staf.STAFUtil;

/**
 * Caches STAX XML files.
 * 
 * @author David Hawkey
 *
 */
public class STAXFileCache
{
    private static final int DEFAULT_CACHE_SIZE = 20;
    
    private static STAXFileCache instance;

    private Map cache;
    private int maxCacheSize;
    private Set localMachineNames;
    
    /**
     * Initializes the STAXCache instance.
     */
    private STAXFileCache()
    {
        maxCacheSize = DEFAULT_CACHE_SIZE;
        cache = Collections.synchronizedMap(new HashMap());
        localMachineNames = new HashSet();
        
        addLocalMachineNames();
    }
    
    /**
     * Adds known local machine name aliases
     */
    private void addLocalMachineNames()
    {
        localMachineNames.add("local");
        localMachineNames.add("localhost");
        localMachineNames.add("127.0.0.1");
        
        try
        {
            InetAddress addr = InetAddress.getLocalHost();
            
            localMachineNames.add(addr.getHostAddress().toLowerCase());
            localMachineNames.add(addr.getHostName().toLowerCase());
            localMachineNames.add(addr.getCanonicalHostName().toLowerCase());
        }
        catch (UnknownHostException e)
        {
            // Do nothing
        }
    }
    
    /**
     * Adds a known local machine name alias.
     * 
     * @param name a local machine name alias such as "localhost".
     */
    public void addLocalMachineName(String name)
    {
        localMachineNames.add(name.toLowerCase());
    }
    
    /**
     * Gets the cache instance.
     * 
     * @return the STAX document cache.
     */
    public static STAXFileCache get()
    {
        if (instance == null)
        {
            instance = new STAXFileCache();
        }
        
        return instance;
    }
    
    /**
     * Checks if the given machine name is a known alias for
     * the local machine.
     * 
     * @param machine the machine name to check.
     * @return whether the machine is known to be an alias to the 
     * local machine.
     */
    public boolean isLocalMachine(String machine)
    {
        return localMachineNames.contains(formatEndpoint(machine));
    }
    
    /**
     * Formats a machine endpoint. This takes a STAF endpoint in the
     * format [&lt;Interface>://]<Logical or Physical Identifier>[@&lt;Port>]
     * and strips off the interface and port leaving only the logical
     * or physical identifier. The identifier will also be converterd to
     * lowercase.
     * 
     * @param machine the machine (STAF endpoint) name to format.
     * @return the logical or physical identifier of the endpoint.
     */
    private String formatEndpoint(String endpoint)
    {
        // Strip the port
        endpoint = STAFUtil.stripPortFromEndpoint(endpoint);

        // Strip the interface
        int index = endpoint.indexOf("://");
        if (index > -1)
        {
            endpoint = endpoint.substring(index + 3);
        }

        endpoint = endpoint.toLowerCase();
        
        return endpoint;
    }
    
    /**
     * Gets a STAX document from the cache. This does not check to see if the
     * cached copy is up-to-date. The method checkCache should first be called
     * (if necessary) to determine if the cached copy exists and is up-to-date.
     * 
     * @param machine the machine containg the XML file.
     * @param filename the STAX XML file name.
     * @return a cached document or null if the document is not in the cache.
     */
    public STAXDocument getDocument(String machine, String filename)
    {
        STAXDocument doc = null;
        
        machine = formatEndpoint(machine);
        
        if (isLocalMachine(machine))
        {
            machine = "local";
        }
        
        FileCacheEntry item = (FileCacheEntry)cache.get(
            new CacheKey(machine, filename));
        
        if (item != null)
        {
            doc = item.getDocument();
            item.hit();
        }
        
        return doc.cloneDocument();
    }
    
    /**
     * Gets the contents of the cache.
     * 
     * @return a collection of STAXCache.CachedSTAX items.
     */
    public List getCacheContents()
    {
    	List contents = null;
    	
    	synchronized(cache)
        {
            contents = new LinkedList(cache.values());
    	}
    	
    	return contents;
    }
    
    /**
     * Clears the contents of the cache.
     * 
     * @return the number of files cleared.
     */
    public int purge()
    {
        int cleared = cache.size();
        cache.clear();
        
        return cleared;
    }
    
    /**
     * Converts a STAX date string to a java Date object.
     * 
     * @param staxDate the STAX date string.
     * @return a java Date object.
     */
    public static Date convertSTAXDate(String staxDate)
    {
        String sYear = staxDate.substring(0,4);
        String sMonth = staxDate.substring(4,4+2);
        String sDay = staxDate.substring(6,6+2);
        String sHour = staxDate.substring(9,9+2);
        String sMinute = staxDate.substring(12,12+2);
        String sSecond = staxDate.substring(15,15+2);
        
        int year = Integer.parseInt(sYear);
        
        // Calendar expects month to be zero-based (0 = January)
        int month = Integer.parseInt(sMonth) - 1;
        int day = Integer.parseInt(sDay);
        int hour = Integer.parseInt(sHour);
        int minute = Integer.parseInt(sMinute);
        int second = Integer.parseInt(sSecond);
        
        Calendar cal = Calendar.getInstance();
        cal.set(year, month, day, hour, minute, second);
        
        // Chop off the milliseconds portion
        // (Should be zero but Calendar sets it to the current system
        // milliseconds)
        long time = ((long)(cal.getTimeInMillis() / 1000))*1000;
        
        return new Date(time);
    }
    
    /**
     * Checks the cache to see if it contains an up-to-date copy of a
     * STAX document.
     * 
     * @param machine the machine that contains the STAX XML.
     * @param filename the filename of the STAX XML.
     * @param date the current last modification date of the file.
     * @return true if the item is cached, false if the item is not cached or
     *             the cached copy is old.
     */
    public boolean checkCache(String machine, String filename, Date date)
    {
    	machine = formatEndpoint(machine);
    	
        if (isLocalMachine(machine))
        {
            machine = "local";
        }
        
        FileCacheEntry item = (FileCacheEntry)cache.get(
            new CacheKey(machine, filename));
        
        if (item != null)
        {
            // Make sure the modification dates are the same, if not the
            // cache is old.

            if (date.equals(item.getModDate()))
            {
                return true;
            }
        }
        
        return false;
    }
    
    /**
     * Adds a STAX document to the cache.
     * 
     * @param name the file name of the STAX xml file.
     * @param doc the parsed STAX document.
     * @param sDate the modification date of the STAX file.
     */
    public void addDocument(String machine, String filename, STAXDocument doc,
                            Date modDate)
    {
    	machine = formatEndpoint(machine);
    	
    	if (isLocalMachine(machine))
        {
            machine = "local";
        }
    	
        FileCacheEntry item = new FileCacheEntry(
            machine, filename, doc.cloneDocument(), modDate);
        
        item.hit();
        
        // Don't add the document if the cache is off

        if (maxCacheSize == 0 )
        {
            return;
        }
        
        // Clean up the cache if it is full

        if ((cache.size() >= maxCacheSize) && (maxCacheSize > 0))
        {
            while(cache.size() >= maxCacheSize)
            {
            	removeOldest();
            }
        }
        
        cache.put(item.getKey(), item);
    }
    
    /**
     * Removes extra items from the cache when the STAX_CACHE_SIZE setting
     * changes.
     */
    public void cleanup()
    {
        // Clean up the cache if it is full

        if ((cache.size() > maxCacheSize) && (maxCacheSize > 0))
        {
            while(cache.size() > maxCacheSize)
            {
            	removeOldest();
            }
        }
    }
    
    /**
     * Removes the oldest entry in the cache
     */
    private void removeOldest()
    {
        // Find the oldest item and remove it

        FileCacheEntry oldest = null;
        
        synchronized(cache)
        {
            Iterator iter = cache.values().iterator();

            if (iter.hasNext())
            {
                oldest = (FileCacheEntry)iter.next();
            }
            
            while(iter.hasNext())
            {
                FileCacheEntry cached = (FileCacheEntry)iter.next();

                if (cached.getLastHitDate().before(oldest.getLastHitDate()))
                {
                    oldest = cached;
                }
            }
        }
        
        cache.remove(oldest.getKey());
    }
    
    /**
     * Sets the maximum number of STAX files to cache.
     * 
     * @param size the maximum STAX files that will be stored in cache.
     */
    public void setMaxCacheSize(int size)
    {
        this.maxCacheSize = size;
        cleanup();
    }
    
    /**
     * Gets the maximum number of STAX files that can be held in cache.
     * 
     * @return the maximum cache size.
     */
    public int getMaxCacheSize()
    {
        return this.maxCacheSize;
    }
    
    private static class CacheKey
    {
        private String machine;
        private String filename;
        
        public CacheKey(String machine, String filename)
        {
            this.machine = machine.toLowerCase();
            this.filename = filename;
        }

        public String getMachine()
        {
            return machine;
        }
        
        public String getFilename()
        {
            return filename;
        }
        
        /* (non-Javadoc)
         * @see java.lang.Object#equals(java.lang.Object)
         */
        public boolean equals(Object o)
        {
            if (this == o)
            {
                return true;
            }
        	
            if (o instanceof CacheKey)
            {
                CacheKey k = (CacheKey)o;
                
                return getMachine().equals(k.getMachine()) &&
                    getFilename().equals(k.getFilename());
            }
            
            return false;
        }

        /* (non-Javadoc)
         * @see java.lang.Object#hashCode()
         */
        public int hashCode()
        {
            return machine.hashCode() + filename.hashCode();
        }
        
    }
    
    /**
     * A Cached STAX document.
     * 
     * @author David Hawkey
     *
     */
    public static class FileCacheEntry
    {
        private String machine;
        private String filename;
        
        private Date date;
        private Date modDate;
        private Date lastHit;
        private int hits;
        private STAXDocument doc;
        
        private CacheKey key;
        
        /**
         * Creates a cached document.
         * 
         * @param doc the document to cache.
         */
        public FileCacheEntry(String machine, String filename,
                              STAXDocument doc, Date modDate)
        {
            this.machine = machine;
            this.filename = filename;
            this.doc = doc;
            this.modDate = modDate;
            this.date = new Date();
            this.lastHit = new Date();
            this.hits = 0;
            
            this.key = new CacheKey(machine, filename);
        }
        
        /**
         * Gets the key used to cache this entry.
         * 
         * @return the cache key for this entry.
         */
        public CacheKey getKey()
        {
            return key;
        }
        
        /**
         * Gets the name of the machine that provided the file.
         * 
         * @return the machine name.
         */
        public String getMachine()
        {
            return machine;
        }
        
        /**
         * Gets the name/path to the file as it exists on the providing
         * machine.
         * 
         * @return the name/path to the file.
         */
        public String getFilename()
        {
            return filename;
        }
        
        /**
         * Gets the STAX document.
         * 
         * @return the cached STAX document.
         */
        public STAXDocument getDocument()
        {
            return doc;
        }
        
        /**
         * Updates the cache hit for this document.
         */
        public void hit()
        {
            lastHit = new Date();
            hits++;
        }
        
        /**
         * Gets the number of cache hits for this document.
         * 
         * @return the number of cache hits.
         */
        public int getHits()
        {
            return hits;
        }
        
        /**
         * Gets the date when the STAX file was last modified.
         * 
         * @return
         */
        public Date getModDate()
        {
            return modDate;
        }
        
        /**
         * Gets last time the document had a cache hit.
         * 
         * @return the last cache hit time.
         */
        public Date getLastHitDate()
        {
            return lastHit;
        }
        
        /**
         * Gets the date when this document was added to the cache.
         * 
         * @return the date the document was added to the cache.
         */
        public Date getAddDate()
        {
            return date;
        }    
        
        /* (non-Javadoc)
         * @see java.lang.Object#equals(java.lang.Object)
         */
        public boolean equals(Object o)
        {
            if (this == o)
            {
                return true;
            }
        	
            if (o instanceof FileCacheEntry)
            {
                FileCacheEntry o1 = (FileCacheEntry)o;
                
                // Two documents are equivalent if they have the same machine, 
                // filename, and mod date.
                boolean equals = 
                    this.getMachine().equalsIgnoreCase(o1.getMachine()) &&
                    this.getFilename().equals(o1.getFilename()) &&
                    this.getModDate().equals(o1.getModDate());
                
                return equals;
            }
            
            return false;
        }
    }

}
