/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002                                              */
/*                                                                           */
/* This software is licensed under the Eclipse Public License (EPL) V1.0.    */
/*****************************************************************************/

package com.ibm.staf.service.stax;
import com.ibm.staf.*;

import java.util.*;

import org.python.core.*;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;

import java.io.File;
import java.io.StringReader;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;

public class STAXImportAction extends STAXActionDefaultImpl
{
    public static String kIgnore = "ignore";
    public static String kError = "error";

    public STAXImportAction()
    { /* Do Nothing */ }

    public STAXImportAction(String machine, String file, String mode)
    { 
        fUnevalMachine = machine;
        fMachine = machine;
        fUnevalFile = file;
        fFile = file;
        fUnevalMode = mode;
        fMode = mode;
    }

    public String getMachine() { return fMachine; } 

    public void setMachine(String machine) 
    {
        fUnevalMachine = machine; 
        fMachine = machine; 
    }
    
    public String getFile() { return fFile; } 

    public void setFile(String file) 
    {
        fUnevalFile = file; 
        fFile = file; 
    }
    
    public String getMode() { return fMode; } 

    public void setMode(String mode)
    {
        fUnevalMode = mode; 
        fMode = mode; 
    }
    
    public String getImportInclude() { return fImportInclude; }

    public void setImportInclude(String importInclude)
    {
        fUnevalImportInclude = importInclude; 
        fImportInclude = importInclude; 
    }

    public String getImportExclude() { return fImportExclude; }

    public void setImportExclude(String importExclude)
    {
        fUnevalImportExclude = importExclude; 
        fImportExclude = importExclude; 
    }
    
    public String getInfo()
    {
        return fMachine + ":" + fFile + ":" + fMode + ":" + fImportExclude;
    }
    
    public String getDetails()
    {
        return "Location:" + fMachine + 
               ";File:" + fFile + 
               ";Mode:" + fMode +
               ";Include:" + fImportInclude +
               ";Exclude:" + fImportExclude;
    }
    
    public String getXMLInfo()
    {
        String xmlInfo = "";
        xmlInfo = "<import machine=\"" + fMachine + "\"" + 
            " file=\"" + fFile + "\"" +
            " mode=\"" + fMode + "\"" +
            ">";
            
        if (!(fImportInclude.equals("")))
        {
            xmlInfo += "\n  <import-include>" + fImportInclude + 
                "</import-include>";
        }
        
        if (!(fImportExclude.equals("")))
        {
            xmlInfo += "\n  <import-exclude>" + fImportExclude + 
                "</import-exclude>";
        }

        xmlInfo += "\n</import>";
               
        return xmlInfo;
    }
    
    public void execute(STAXThread thread)
    {
            /* STAXResult will be set to a list containing:
            
               - None, or a list containing a STAXImportError object and a text
                 string with details about the error.
             
               - A list of the successfully imported functions that were 
                 requested to be imported.
                 
               - A list of the successfully imported functions that were 
                 required by other functions.
                 
               - A list of the functions that were requested to be imported 
                 but already existed (so they were not imported).
                 
               - A list of the functions that were required by other functions
                 but already existsed (so they were not imported).
                 
               - A list of the functions there were not requested to be
                 imported and were not required by other functions.
                 
               - A list of functions requested to be imported that were not
                 found.
           */            
    
        fThread = thread;
        
        fFunctionListDoesNotExist = new ArrayList();
        fFunctionListNotRequestedNotRequired = new ArrayList();
        fFunctionListImportedRequested = new ArrayList();
        fFunctionListImportedRequired = new ArrayList();
        fFunctionListExistingRequested = new ArrayList();
        fFunctionListExistingRequired = new ArrayList();
        
        String evalElem = getElement();
        String evalAttr = "machine";

        try
        {
            String STAXResult = "";

            fMachine = fThread.pyStringEval(fUnevalMachine);

            evalAttr = "file";
            fFile = fThread.pyStringEval(fUnevalFile);

            evalAttr = "mode";
            fMode = fThread.pyStringEval(fUnevalMode);
            
            if (!fMode.equalsIgnoreCase("error") &&
                !fMode.equalsIgnoreCase("ignore"))
            {
                // Invalid mode
                fThread.popAction();

                String errorData = "[STAXImportModeError, r'" + 
                    "Invalid import mode: " + fMode + "']";

                PyObject signalData = fThread.pyObjectEval(errorData);
                fThread.pySetVar("STAXSignalData", signalData);

                setElementInfo(new STAXElementInfo(
                    evalElem, evalAttr,
                    "Invalid import mode: " + fMode +
                    "\nImport mode must be 'error' or 'ignore'."));

                fThread.setSignalMsgVar(
                    "STAXImportErrorMsg", STAXUtil.formatErrorMessage(this));

                fThread.raiseSignal("STAXImportError");

                return;
            }

            boolean emptyImportList = false;
            boolean emptyExcludeList = false;
            Object[] array;
            Vector importList = new Vector();
            Vector excludeList = new Vector();
           
            evalElem = "import-include";
            evalAttr = STAXElementInfo.NO_ATTRIBUTE_NAME;

            if (fUnevalImportInclude.equals(""))
            {
                emptyImportList = true;
            }
            else
            {
                List list = fThread.pyListEval(fUnevalImportInclude);
                importList = new Vector(list);
            }

            evalElem = "import-exclude";

            if (fUnevalImportExclude.equals(""))
            {
                emptyExcludeList = true;
            }
            else
            {
                List list = fThread.pyListEval(fUnevalImportExclude);
                excludeList = new Vector(list);
            }
             
            evalElem = getElement();

            STAXParser parser = new STAXParser(fThread.getJob().getSTAX());
            
            Date dLastModified = null;
            
            STAXJob job = null;
            
            // If caching is enabled, find the modification date

            if (fThread.getJob().getSTAX().getFileCaching())
            {
                if (STAXFileCache.get().isLocalMachine(fMachine))
                {
                    File file = new File(fFile);
                    
                    // Make sure the file exists

                    if (file.exists())
                    {
                        long lastModified = file.lastModified();

                        if (lastModified > 0)
                        {
                            // Chop off the milliseconds because some
                            // systems don't report mod time to milliseconds

                            lastModified = ((long)(lastModified/1000))*1000;

                            dLastModified = new Date(lastModified);
                        }
                    }
                }
                
                if (dLastModified == null)
                {
                    // Find the remote file mod time using STAF

                    STAFResult entryResult = fThread.getJob().submitSync(
                        fMachine, "FS", "GET ENTRY " + fFile + " MODTIME");

                    if (entryResult.rc == 0)
                    {
                        String modDate = entryResult.result;
                        dLastModified = STAXFileCache.convertSTAXDate(
                            modDate);
                    }
                }
                
                // Check for an up-to-date file in the cache

                if ((dLastModified != null) &&
                    STAXFileCache.get().checkCache(
                        fMachine, fFile, dLastModified))
                {
                    // Get the doc from cache

                    STAXDocument doc = STAXFileCache.get().getDocument(
                        fMachine, fFile);
                    
                    job = new STAXJob(fThread.getJob().getSTAX(), doc);
                }
            }
            
            // If the file was not in cache, then retrieve it using STAF

            if (job == null)
            {
                STAFResult getResult = fThread.getJob().submitSync(
                    fMachine, "FS", "GET FILE " + fFile);
                               
                if (getResult.rc != 0)           
                {
                    String errorData = "";
                    
                    if (getResult.rc == STAFResult.NoPathToMachine)
                    {
                        errorData = "[STAXNoResponseFromMachine, r'" + 
                            "RC=" + getResult.rc + " Result=" + 
                            getResult.result + "']";
                    }
                    else
                    {
                        errorData = "[STAXFileCopyError, r'" +
                            "RC=" + getResult.rc + " Result=" + 
                            getResult.result + "']";
                    }
    
                    fThread.popAction();
                    
                    if (fMode.equals(kError))
                    {
                        PyObject signalData = 
                            fThread.pyObjectEval(errorData);
                        fThread.pySetVar("STAXSignalData", signalData);

                        setElementInfo(new STAXElementInfo(
                            evalElem, evalAttr,
                            "FS GET request to import file " + fFile +
                            " from machine " + fMachine +
                            " failed.\nRC: " + getResult.rc +
                            ", Result: " + getResult.result));

                        fThread.setSignalMsgVar(
                            "STAXImportErrorMsg",
                            STAXUtil.formatErrorMessage(this));

                        fThread.raiseSignal("STAXImportError");
                    }
                    else
                    {
                        // just set STAXResult with an error as first element
                        String result = "[" + errorData + ",[],[],[],[],[],[]]";
                        PyObject resultObject = fThread.pyObjectEval(result);
                        fThread.pySetVar("STAXResult", resultObject);
                    }
                     
                    return;
                }
                
                // Parse the XML document
                job = parser.parse(getResult.result, fFile, fMachine);
                
                // Add the XML document to the cache
                if (fThread.getJob().getSTAX().getFileCaching() &&
                    (dLastModified != null))
                {
                    STAXFileCache.get().addDocument(
                        fMachine, fFile, job.getSTAXDocument(),
                        dLastModified);
                }
            }
                
            fFunctionMap = job.getSTAXDocument().getFunctionMap();
            Set functionSet = fFunctionMap.keySet();

            fThread.pyExec("import re");

            // iterator through the imported xml file's functions
            Iterator functionIterator = functionSet.iterator();

            while (functionIterator.hasNext())
            {
                String nextFunction = (String)functionIterator.next();
                
                boolean includeMatch = false;
                boolean excludeMatch = false;

                if (emptyImportList)
                {
                    checkForMatch(fFunctionMap, nextFunction, excludeList);
                }
                else
                {
                    boolean included = false;
                    
                    for (int im = 0; im < importList.size(); im++)
                    {
                        includeMatch = grepMatch(nextFunction,
                            (importList.elementAt(im)).toString());
                        
                        if (includeMatch)
                        {
                            included = checkForMatch(fFunctionMap, 
                                nextFunction, excludeList);
                        }
                    }
                    
                    if (!included && !(fFunctionListImportedRequired.
                        contains(nextFunction)))
                    {
                        if (!(fFunctionListNotRequestedNotRequired.
                            contains(nextFunction)))
                        {
                            fFunctionListNotRequestedNotRequired.add(
                            nextFunction); 
                        }
                    }
                }
            }
            
            // process required functions
            for (int rf = 0; rf < fRequiredFunctions.size(); rf++)
            {
                addRequiredFunctions((String)fRequiredFunctions.elementAt(rf));
            }
            
            // handle any non-"*" functions that were not found
            for (int im = 0; im < importList.size(); im++)
            {
                if ((importList.elementAt(im).toString()).indexOf("*") == -1)
                {
                    if (!(fFunctionMap.containsKey(importList.elementAt(im).toString()))
                        && !(excludeList.contains(importList.elementAt(im))))
                    {
                        fFunctionListDoesNotExist.add(
                            importList.elementAt(im).toString());
                    }
                }
            }
            
            STAXResult += "[";
            STAXResult += "None";
            STAXResult += ",";
            STAXResult += createListString(fFunctionListImportedRequested) + 
                ",";
            STAXResult += createListString(fFunctionListImportedRequired) + 
                ",";
            STAXResult += createListString(fFunctionListExistingRequested) + 
                ",";
            STAXResult += createListString(fFunctionListExistingRequired) + 
                ",";
            STAXResult += createListString(
                fFunctionListNotRequestedNotRequired) + ",";
            STAXResult += createListString(fFunctionListDoesNotExist);
            STAXResult += "]";
            
            PyObject resultObject = fThread.pyObjectEval(STAXResult);
            fThread.pySetVar("STAXResult", resultObject);
        }
        catch (STAXPythonEvaluationException ex)
        {
            fThread.popAction();

            setElementInfo(new STAXElementInfo(evalElem, evalAttr));

            fThread.setSignalMsgVar(
                "STAXPythonEvalMsg", STAXUtil.formatErrorMessage(this), ex);

            fThread.raiseSignal("STAXPythonEvaluationError");

            return;
        }
        catch (STAXXMLParseException ex)
        {
            handleParserException(ex);   
            return;
        }
        catch (STAXException ex)
        {
            handleParserException(ex);   
            return;
        }
        catch (SAXNotRecognizedException ex)
        {
            handleParserException(ex);   
            return;
        }
        catch (SAXNotSupportedException ex)
        {
            handleParserException(ex);   
            return;
        }
        
        // do something 
         
        fThread.popAction();
    }

    public void handleParserException(Exception ex)
    {    
        fThread.popAction();

        // Create a PyList containing "STAXXMLParseError" and the
        // parser exception string. 

        PyList errorData = new PyList();
        errorData.append(new PyString("STAXXMLParseError"));
        errorData.append(new PyString(fFile + ": " + ex.toString()));
        
        if (fMode.equals(kError))
        {
            String errorDataStr = "[STAXXMLParseError, r'" + fFile + ": " +
                ex.toString().replace('\n', ' ') + "']";

            try
            {
                PyObject signalData = fThread.pyObjectEval(errorDataStr);
                fThread.pySetVar("STAXSignalData", signalData);
            }
            catch (STAXPythonEvaluationException e)
            {
                // XXX: Should never happen
                System.out.println("fThread.pyObjectEval failed for: " +
                                   errorDataStr + "\n" + e.toString());
                fThread.pySetVar("STAXSignalData", errorData);
            }

            String msg = "File: " + fFile + ", Machine: " + fMachine + "\n" +
                ex.toString();

            fThread.setSignalMsgVar("STAXImportErrorMsg", msg);
            fThread.raiseSignal("STAXImportError");
        }
        else
        {
            // Create STAXResult as a list with errorData as its first
            // element and with empty lists for its next 6 elements:
            // STAXResult = [ errorData,[],[],[],[],[],[] ]

            PyList resultList = new PyList();
            PyList emptyList = new PyList();

            resultList.append(errorData);

            for (int i = 1; i < 7; ++i)
                resultList.append(emptyList);
           
            fThread.pySetVar("STAXResult", resultList);
        }
    }    

    public String createListString(ArrayList list)
    {
        String listString = "[";
        
        Iterator listIter = list.iterator();
        
        while (listIter.hasNext())
        {
            if (listString.equals("["))
            {
                listString += "'" + listIter.next() + "'";
            }
            else
            {
                listString += ",'" + listIter.next() + "'";
            }
        }

        listString += "]";
        
        return listString;
    }

    public void addRequiredFunctions(String currentFunction)
    {
        if (!(fThread.getJob().functionExists(currentFunction)))
        {
            STAXFunctionAction function = 
                (STAXFunctionAction)fFunctionMap.get(currentFunction);        

            fThread.getJob().addFunction(function);
           
            fFunctionListImportedRequired.add(currentFunction);
            fFunctionListNotRequestedNotRequired.remove(currentFunction);

            StringTokenizer requiredFunctions = 
                new StringTokenizer(function.getRequires(), " ");
            
            while (requiredFunctions.hasMoreElements())
            {
                addRequiredFunctions((String)requiredFunctions.nextElement());
            }
        }
        else
        {
            if (!(fFunctionListExistingRequired.contains(currentFunction)) &&
                !(fFunctionListImportedRequired.contains(currentFunction)))
            {
                fFunctionListExistingRequired.add(currentFunction);
            }
        }
    }

    public void handleCondition(STAXThread thread, STAXCondition cond)
    {
        thread.popAction();
    }

    public STAXAction cloneAction()
    {
        STAXImportAction clone = new STAXImportAction();

        clone.setElement(getElement());
        clone.setLineNumberMap(getLineNumberMap());
        clone.setXmlFile(getXmlFile());
        clone.setXmlMachine(getXmlMachine());

        clone.fUnevalMachine = fUnevalMachine;
        clone.fMachine = fMachine;
        clone.fUnevalFile = fUnevalFile;
        clone.fFile = fFile;
        clone.fUnevalMode = fUnevalMode;
        clone.fMode = fMode;
        clone.fUnevalImportInclude = fUnevalImportInclude;
        clone.fUnevalImportExclude = fUnevalImportExclude;
        clone.fImportInclude = fImportInclude;
        clone.fImportExclude = fImportExclude;

        return clone;
    }
    
    private boolean grepMatch(String targetString, String matchString)
        throws STAXPythonEvaluationException
    {
        return fThread.pyBoolEval("re.match('" + matchString + 
           "', '" + targetString + "')");
    }
    
    private boolean checkForMatch(HashMap functionMap, String function,
                                  Vector excludeList) 
                                  throws STAXPythonEvaluationException
    {
        boolean excludeMatch = false;
        boolean included = false;
        
        for (int ex = 0; ex < excludeList.size(); ex++)
        {
            if (grepMatch(function, excludeList.elementAt(ex).toString()))
            {
                excludeMatch = true;
            }
        }
                    
        if (!excludeMatch)
        {
            handleFunctionImport(functionMap, function);
            
            included = true;
        }
        else
        {
            if (!(fFunctionListNotRequestedNotRequired.contains(function)))
            {
                fFunctionListNotRequestedNotRequired.add(function);
            }
        }
        
        return included;
    }
    
    private void handleFunctionImport(HashMap functionMap, 
                                      String functionToImport)
    {
        STAXFunctionAction function = 
            (STAXFunctionAction)functionMap.get(functionToImport);
                    
        StringTokenizer requiredFunctions = 
            new StringTokenizer(function.getRequires(), " ");
            
        while (requiredFunctions.hasMoreElements())
        {
            fRequiredFunctions.add(requiredFunctions.nextElement());
        }
                    
        if (!(fThread.getJob().functionExists(functionToImport)))
        {
            // add the function

            fThread.getJob().addFunction(function);
            
            fFunctionListImportedRequested.add(functionToImport);
        }
        else
        {
            // function already exists or has already been imported
            
            fFunctionListExistingRequested.add(functionToImport);
        }
    }

    STAXThread fThread = null;    
    private String fUnevalMachine = new String();
    private String fMachine = new String();    
    private String fUnevalFile = new String();
    private String fFile = new String();
    private String fUnevalMode = new String();
    private String fMode = new String();
    private String fUnevalImportInclude = new String();
    private String fUnevalImportExclude = new String();
    private String fImportInclude = new String();
    private String fImportExclude = new String();
    
    private ArrayList fFunctionListDoesNotExist;
    private ArrayList fFunctionListNotRequestedNotRequired;
    private ArrayList fFunctionListImportedRequested;
    private ArrayList fFunctionListImportedRequired;
    private ArrayList fFunctionListExistingRequested;
    private ArrayList fFunctionListExistingRequired;

    private Vector fRequiredFunctions = new Vector();
    private HashMap fFunctionMap;
}