/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002                                              */
/*                                                                           */
/* This software is licensed under the Eclipse Public License (EPL) V1.0.    */
/*****************************************************************************/

package com.ibm.staf.service.stax;
import java.util.List;
import java.util.HashMap;
import java.util.Iterator;

public class STAXParallelIterateAction extends STAXActionDefaultImpl
                                       implements STAXThreadCompleteListener
{
    static final int INIT = 0;
    static final int WAIT_THREADS = 1;
    static final int THREADS_COMPLETE = 2;
    static final int COMPLETE = 3;

    static final String INIT_STRING = "INIT";
    static final String WAIT_THREADS_STRING = "WAIT_THREADS";
    static final String THREADS_COMPLETE_STRING = "THREADS_COMPLETE";
    static final String COMPLETE_STRING = "COMPLETE";
    static final String STATE_UNKNOWN_STRING = "UNKNOWN";

    public STAXParallelIterateAction()
    { /* Do Nothing */ }

    public STAXParallelIterateAction(String itemvar, String in, 
                                     String indexvar, STAXAction action)
    {
        fItemvar = itemvar;
        fIn = in;
        fIndexvar = indexvar; 
        fAction = action;
    }

    public void setItemVar(String itemvar) { fItemvar = itemvar; }
    public void setIn(String in) { fIn = in; }
    public void setIndexVar(String indexvar) { fIndexvar = indexvar; }
    public void setAction(STAXAction action) { fAction = action; }

    public String getStateAsString()
    {
        switch (fState)
        {
            case INIT:
                return INIT_STRING;
            case WAIT_THREADS:
                return WAIT_THREADS_STRING;
            case THREADS_COMPLETE:
                return THREADS_COMPLETE_STRING;
            case COMPLETE:
                return COMPLETE_STRING;
            default:
                return STATE_UNKNOWN_STRING;
        }
    }

    public String getXMLInfo()
    {
        if (fIndexvar.equals(""))
            return "<paralleliterate var=\"" + fItemvar + "\" in=\"" + 
                   fIn + "\">";
        else
            return "<paralleliterate var=\"" + fItemvar + "\" in=\"" + 
                   fIn + "\" indexvar=\"" + fIndexvar + "\">";
    }

    // Used to generate the Call Stack information for this element

    public String getInfo()
    {
        // Format:
        //   <listSize> <state> <list>
        //
        //  where:
        //    <listSize> is the size of the list (or 0 if not available yet)
        //    <state>    is the string version of the current state
        //    <list>     is the unevaluated list value
        //
        //  The total length of the message is limited to 40 characters.
        //  If it exceeds 40 characters, "..." is appended.

        String info =  fListSize + " " + getStateAsString() + " " + fIn;

        if (info.length() > 40)
            info += info.substring(0, 40) + "...";

        return info;
    }

    public String getDetails()
    {
        return "ListSize:" + fListSize + 
               ";State:" + getStateAsString() +
               ";Itemvar:" + fItemvar +
               ";In:" + fIn +
               ";Indexvar:" + fIndexvar +
               ";Action:" + fAction +
               ";HoldThreadCondition:" + fHoldCondition +
               ";HardHoldThreadCondition:" + fHardHoldCondition + 
               ";ThreadMap:" + fThreadMap +
               ";List:" + fList;
    }
    
    public void execute(STAXThread thread)
    {
        String evalElem = getElement();
        String evalAttr = "in";

        try
        {
            if (fState == INIT)
            {
                // Create a Python list/tuple from fIn and then extract it
                // into a Java List.

                fList = thread.pyListEval(fIn);
                fListSize = fList.size();

                if (fListSize == 0)
                {
                    // Nothing in list, no threads to start; raise a signal.

                    fState = THREADS_COMPLETE;
                    thread.popAction();

                    setElementInfo(new STAXElementInfo(evalElem, evalAttr));

                    thread.setSignalMsgVar(
                        "STAXEmptyListMsg",
                        STAXUtil.formatErrorMessage(this));

                    thread.raiseSignal("STAXEmptyList");

                    return;
                }
                
                synchronized (fThreadMap)
                {
                    for (int i = 0; i < fListSize; i++)
                    {
                        // Setup up a new thread for each iteration
                        STAXThread childThread = thread.createChildThread();
  
                        // Assign indexvar Jython variable, if provided

                        if (fIndexvar != null && !fIndexvar.equals(""))
                        {
                            evalAttr = "indexvar";
                            childThread.pySetVar(fIndexvar, new Integer(i));
                        }

                        // Assign itemvar Jython variable

                        evalAttr = "var";
                        Object itemValue = (Object)fList.get(i);

                        childThread.pySetVar(fItemvar, itemValue);
                        fThreadMap.put(new Integer(
                            childThread.getThreadNumber()), childThread);
                        childThread.addCompletionNotifiee(this);
                        childThread.pushAction(fAction.cloneAction());
                        childThread.schedule();
                    }
                        
                    fState = WAIT_THREADS;
                    thread.addCondition(fHoldCondition);
                }
          
                return;
            }
            else if (fState == THREADS_COMPLETE)
            {
                fState = COMPLETE;
                thread.popAction();
            }   
        }
        catch (STAXPythonEvaluationException e)
        {
            fState = COMPLETE;
            thread.popAction();

            setElementInfo(new STAXElementInfo(evalElem, evalAttr));

            thread.setSignalMsgVar(
                "STAXPythonEvalMsg",
                STAXUtil.formatErrorMessage(this), e);

            thread.raiseSignal("STAXPythonEvaluationError");
        }
    }

    public void handleCondition(STAXThread thread, STAXCondition cond)
    {
        synchronized (fThreadMap)
        {
            if (!fThreadMap.isEmpty())
            {
                // Add a hard hold, so we can wait until our children terminate.
                // This is removed by threadComplete().

                thread.addCondition(fHardHoldCondition);

                // Now, iterate of our children and tell them to terminate

                Iterator iter = fThreadMap.values().iterator();

                while (iter.hasNext())
                {
                    STAXThread childThread = (STAXThread)iter.next();
                    childThread.terminate(
                        STAXThread.THREAD_END_STOPPED_BY_PARENT);
                }
            }
        }

        fState = COMPLETE;  
        thread.popAction();
    }

    public STAXAction cloneAction()
    {
        STAXParallelIterateAction clone = new STAXParallelIterateAction();

        clone.setElement(getElement());
        clone.setLineNumberMap(getLineNumberMap());
        clone.setXmlFile(getXmlFile());
        clone.setXmlMachine(getXmlMachine());

        clone.fItemvar = fItemvar;
        clone.fIn = fIn;
        clone.fIndexvar = fIndexvar;
        clone.fAction = fAction;

        return clone;
    }

    // STAXThreadCompleteListener method

    public void threadComplete(STAXThread thread, int endCode)
    {
        synchronized (fThreadMap)
        {
            fThreadMap.remove(new Integer(thread.getThreadNumber()));

            if (fThreadMap.isEmpty())
            {
                fState = THREADS_COMPLETE;

                // thread.getParentThread() should be the thread we are
                // running on.

                thread.getParentThread().removeCondition(fHoldCondition);
                thread.getParentThread().removeCondition(fHardHoldCondition);
                thread.getParentThread().schedule();
            }
        }
    }

    private STAXHoldThreadCondition fHoldCondition = 
        new STAXHoldThreadCondition("ParallelIterate");
    private STAXHardHoldThreadCondition fHardHoldCondition =
        new STAXHardHoldThreadCondition("ParallelIterate");
    private int fState = INIT;
    private HashMap fThreadMap = new HashMap();
    private List fList = null;
    private int fListSize = 0;

    private String fItemvar = new String();
    private String fIn = new String();
    private String fIndexvar = new String();
    private STAXAction fAction = null;
}
