/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001, 2004                                        */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAF_iostream.h"
#include "STAF_fstream.h"
#include "STAFString.h"
#include "STAFProc.h"
#include "STAFProcUtil.h"
#include "STAFConfig.h"
#include "STAFUtil.h"
#include "STAFConnectionManager.h"
#include "STAFCommandParser.h"
#include "STAFException.h"
#include "STAFDelegatedService.h"
#include "STAFExternalService.h"
#include "STAFNotificationList.h"
#include "STAFTrustManager.h"
#include "STAFDiagManager.h"
#include "STAFServiceManager.h"
#include "STAFThreadManager.h"
#include "STAFProcessService.h"
#include "STAFTrace.h"
#include "STAFTraceService.h"

unsigned int handleMachineNicknameConfig(STAFString line);
unsigned int handleInterfaceConfig(STAFString line);
unsigned int handleTrustConfig(STAFString line);
unsigned int handleNotifyConfig(STAFString line);
unsigned int handleTraceConfig(STAFString line);
unsigned int handleServiceConfig(STAFString line);
unsigned int handleServiceLoaderConfig(STAFString line);
unsigned int handleAuthenticatorConfig(STAFString line);
unsigned int handleSetConfig(STAFString line);

unsigned int fSLSNum = 1;
const STAFVariablePool *varPoolList[] = { 0, 0 };
unsigned int varPoolListSize = 2;
const unsigned int maxLineLength = 2048;

unsigned int readConfigFile(const STAFString &filename)
{
    // Init the variable pool list
    varPoolList[0] = *gSharedVariablePoolPtr;
    varPoolList[1] = *gGlobalVariablePoolPtr;

    unsigned int rc = 0;

    ifstream config(filename.toCurrentCodePage()->buffer());

    if (!config)
    {
        cout << "Could not find config file " << filename << endl;
        return 1;
    }

    STAFString whiteSpace("\x20\x09\x0A\x0D");
    STAFString line;
    char configLine[maxLineLength];

    for(; config.good() && !rc;)
    {
        configLine[0] = 0;
        config.getline(configLine, maxLineLength);

        if (!config.good())
        {
            if (!config.eof())
            {
                cout << "Line in configuration file exceeds maximum length of "
                     << STAFString(maxLineLength) << " characters" << endl;
                rc = 1;
                break;
            }

            if (configLine == "") break;
        }

        STAFString thisLine(configLine);

        if ((thisLine == "") ||
            (thisLine.findFirstNotOf(whiteSpace) == STAFString::kNPos) ||
            (thisLine.subString(0, 1) == kUTF8_POUND))
        {
            if (line.length() != 0)
            {
                cout << "Invalid continuation line, " << thisLine << endl;
                rc = 1;
            }

            continue;
        }
        else if (thisLine.subString(thisLine.length() - 1, 1) == kUTF8_BSLASH)
        {
            line += thisLine.subString(0,
                    thisLine.length() - 1).strip(STAFString::kFront);
            continue;
        }
        else
        {
            line += thisLine.strip(STAFString::kFront);
        }

        STAFString firstLowerWord = line.subWord(0, 1).lowerCase();

        if (firstLowerWord == "machinenickname")
            rc = handleMachineNicknameConfig(line);
        else if (firstLowerWord == "interface")
            rc = handleInterfaceConfig(line);
        else if (firstLowerWord == "trust")
            rc = handleTrustConfig(line);
        else if (firstLowerWord == "notify")
            rc = handleNotifyConfig(line);
        else if (firstLowerWord == "trace")
            rc = handleTraceConfig(line);
        else if (firstLowerWord == "service")
            rc = handleServiceConfig(line);
        else if (firstLowerWord == "serviceloader")
            rc = handleServiceLoaderConfig(line);
        else if (firstLowerWord == "authenticator")
            rc = handleAuthenticatorConfig(line);
        else if (firstLowerWord == "set")
            rc = handleSetConfig(line);
        else
        {
            cout << "Unknown config command, " << line << endl;
            rc = 1;
        }

        // Reset line for the next statement
        line = STAFString();
    }

    return rc;
}

unsigned int handleTrustConfig(STAFString line)
{
    STAFCommandParser fTrustParser;

    // trust config options

    fTrustParser.addOption("TRUST",   1,
        STAFCommandParser::kValueNotAllowed);
    fTrustParser.addOption("LEVEL",   1,
        STAFCommandParser::kValueRequired);
    fTrustParser.addOption("MACHINE", 0,
        STAFCommandParser::kValueRequired);
    fTrustParser.addOption("USER", 0,
        STAFCommandParser::kValueRequired);
    fTrustParser.addOption("DEFAULT", 1,
        STAFCommandParser::kValueNotAllowed);

    // trust config groups

    fTrustParser.addOptionGroup("LEVEL",                1, 1);
    fTrustParser.addOptionGroup("MACHINE USER DEFAULT", 1, 1);

    STAFCommandParseResultPtr parsedResult = fTrustParser.parse(line);

    if (parsedResult->rc != kSTAFOk)
    {
        cout << "Error on TRUST definition, " << line << endl
             << parsedResult->errorBuffer << endl;

        return 1;
    }

    STAFString errorBuffer;
    STAFString levelString;
    STAFRC_t rc = RESOLVE_STRING_OPTION("LEVEL", levelString);

    if (rc != kSTAFOk)
    {
        cout << "Error on TRUST definition, " << line << endl
             << "Error resolving " << parsedResult->optionValue("LEVEL")
             << ", RC: " << rc  << endl;
        return 1;
    }

    if (!levelString.isDigits())
    {
        cout << "Error on TRUST definition, " << line << endl
             << "The <Level> must be a non-negative integer" << endl;
        return 1;
    }

    unsigned int level = levelString.asUInt();

    if (parsedResult->optionTimes("DEFAULT") != 0)
    {
        rc = gTrustManagerPtr->setDefaultTrusteeLevel(level);

        if (rc != kSTAFOk)
        {
            cout << "Error setting DEFAULT TRUST definition, " << line
                 << endl << "Error code: " << rc << endl;
            return 1;
        }
    }
    else if (parsedResult->optionTimes("MACHINE") != 0)
    {
        for(int i = 1; i <= parsedResult->optionTimes("MACHINE"); ++i)
        {
            STAFString machine;
            rc = RESOLVE_INDEXED_STRING_OPTION("MACHINE", i, machine);

            if (rc != kSTAFOk)
            {
                cout << "Error on TRUST definition, " << line << endl
                     << "Error resolving "
                     << parsedResult->optionValue("MACHINE", i) << ", RC: "
                     << rc << endl;
                return 1;
            }

            rc = gTrustManagerPtr->setMachineTrusteeLevel(machine, level);

            if (rc != kSTAFOk)
            {
                cout << "Error adding TRUST definition, " << line
                     << endl << "Error code: " << rc << endl;
                return 1;
            }
        }
    }
    else
    {
        for(int i = 1; i <= parsedResult->optionTimes("USER"); ++i)
        {
            STAFString user;
            RESOLVE_INDEXED_STRING_OPTION("USER", i, user);

            if (rc != kSTAFOk)
            {
                cout << "Error on TRUST definition, " << line << endl
                     << "Error resolving "
                     << parsedResult->optionValue("USER", i) << ", RC: "
                     << rc << endl;
                return 1;
            }

            // XXX: Need to specify whether level is for machine or user
            rc = gTrustManagerPtr->setUserTrusteeLevel(user, level);

            if (rc != kSTAFOk)
            {
                cout << "Error adding TRUST definition, " << line
                     << endl << "Error code: " << rc << endl;
                return 1;
            }
        }
    }

    return 0;
}


unsigned int handleInterfaceConfig(STAFString line)
{
    STAFCommandParser fInterfaceParser;

    // interface config options

    fInterfaceParser.addOption("INTERFACE", 1,
        STAFCommandParser::kValueRequired);
    fInterfaceParser.addOption("LIBRARY",   1,
        STAFCommandParser::kValueRequired);
    fInterfaceParser.addOption("OPTION",    0,
        STAFCommandParser::kValueRequired);

    // interface config groups

    fInterfaceParser.addOptionGroup("INTERFACE", 1, 1);
    fInterfaceParser.addOptionGroup("LIBRARY",   1, 1);

    STAFCommandParseResultPtr parsedResult = fInterfaceParser.parse(line);

    if (parsedResult->rc != kSTAFOk)
    {
        cout << "Error on INTERFACE definition, " << line << endl
             << parsedResult->errorBuffer << endl;
        return 1;
    }

    STAFString errorBuffer;
    STAFString name;
    STAFRC_t rc = RESOLVE_STRING_OPTION("INTERFACE", name);

    if (rc != kSTAFOk)
    {
        cout << "Error on INTERFACE definition, " << line << endl
             << "Error resolving INTERFACE name "
             << parsedResult->optionValue("INTERFACE") << ", RC: " << rc
             << endl;
        return 1;
    }

    if (name.toLowerCase() == "local")
    {
        cout << "Error on INTERFACE definition, " << line << endl
             << "Interface name " << name
             << " is reserved for the local IPC interface" << endl;
        return 1;
    }

    STAFString library;
    RESOLVE_STRING_OPTION("LIBRARY", library);

    if (rc != kSTAFOk)
    {
        cout << "Error on INTERFACE definition, " << line << endl
             << "Error resolving LIBRARY "
             << parsedResult->optionValue("LIBRARY") << ", RC: " << rc << endl;
        return 1;
    }

    STAFConnectionManager::ConnectionProviderOptionList optionList;

    for (int i = 1; i <= parsedResult->optionTimes("OPTION"); ++i)
    {
        STAFString option;
        RESOLVE_INDEXED_STRING_OPTION("OPTION", i, option);

        if (rc != kSTAFOk)
        {
            cout << "Error on INTERFACE definition, " << line << endl
                 << "Error resolving OPTION"
                 << parsedResult->optionValue("OPTION", i)<< ", RC: " << rc
                 << endl;
            return 1;
        }

        optionList.push_back(option);
    }

    rc = gConnectionManagerPtr->addConnectionProvider(name, library, optionList,
                                                      errorBuffer);

    if (rc != kSTAFOk)
    {
        cout << "Error on INTERFACE definition, " << line << endl
             << "Error creating interface, RC: " << rc << ", Reason: "
             << errorBuffer << endl;
        return 1;
    }

    return 0;
}

unsigned int handleMachineNicknameConfig(STAFString line)
{
    STAFCommandParser fMachineNicknameParser;

    // machine config options

    fMachineNicknameParser.addOption("MACHINENICKNAME", 1,
        STAFCommandParser::kValueRequired);

    STAFCommandParseResultPtr parsedResult = fMachineNicknameParser.parse(line);

    if (parsedResult->rc != kSTAFOk)
    {
        cout << "Error on MACHINENICKNAME definition, " << line << endl
             << parsedResult->errorBuffer << endl;
        return 1;
    }
    
    STAFString errorBuffer;
    STAFString machineNickname;
    STAFRC_t rc = RESOLVE_STRING_OPTION("MACHINENICKNAME", machineNickname);

    if (rc != kSTAFOk)
    {
        cout << "Error on MACHINENICKNAME definition, " << line << endl
             << "Error resolving "
             << parsedResult->optionValue("MACHINENICKNAME")
             << ", RC: " << rc << endl;
        return 1;
    }

    *gMachineNicknamePtr = machineNickname;

    return 0;
}


unsigned int handleNotifyConfig(STAFString line)
{
    STAFCommandParser fNotifyParser;

    // notify config options

    fNotifyParser.addOption("NOTIFY",     1,
        STAFCommandParser::kValueNotAllowed);
    fNotifyParser.addOption("ONSTART",    1,
        STAFCommandParser::kValueNotAllowed);
    fNotifyParser.addOption("ONSHUTDOWN", 1,
        STAFCommandParser::kValueNotAllowed);
    fNotifyParser.addOption("NAME",       1,
        STAFCommandParser::kValueRequired);
    fNotifyParser.addOption("HANDLE",     1,
        STAFCommandParser::kValueRequired);
    fNotifyParser.addOption("MACHINE",    1,
        STAFCommandParser::kValueRequired);
    fNotifyParser.addOption("PRIORITY",   1,
        STAFCommandParser::kValueRequired);

    // notify config groups

    fNotifyParser.addOptionGroup("NAME HANDLE",        1, 1);
    fNotifyParser.addOptionGroup("ONSTART ONSHUTDOWN", 1, 1);
    fNotifyParser.addOptionGroup("MACHINE",            1, 1);
    fNotifyParser.addOptionGroup("NOTIFY",             1, 1);

    STAFCommandParseResultPtr parsedResult = fNotifyParser.parse(line);

    if (parsedResult->rc != kSTAFOk)
    {
        cout << "Error on NOTIFY definition, " << line << endl
             << parsedResult->errorBuffer << endl;
        return 1;
    }

    unsigned int onStart = parsedResult->optionTimes("ONSTART");
    unsigned int notifyName = parsedResult->optionTimes("NAME");
    unsigned int priority = 5;
    unsigned int handle = 0;
    STAFString name;
    STAFString machine;
    STAFString errorBuffer;
    STAFRC_t rc = RESOLVE_OPTIONAL_UINT_OPTION("PRIORITY", priority);

    if (!rc) RESOLVE_OPTIONAL_UINT_OPTION("HANDLE", handle);
    if (!rc) RESOLVE_OPTIONAL_STRING_OPTION("NAME", name);
    if (!rc) RESOLVE_STRING_OPTION("MACHINE", machine);

    if (rc != kSTAFOk)
    {
        cout << "Error on NOTIFY definition, " << line << endl
             << "Error with, " << errorBuffer << ", RC: " << rc << endl;
        return 1;
    }

    STAFNotificationList *notificationList = gNotifyOnStartPtr;

    if (!onStart) notificationList = gNotifyOnShutdownPtr;

    if (notifyName) rc = notificationList->reg(machine, name, priority);
    else rc = notificationList->reg(machine, handle, priority);

    return rc;
}


unsigned int handleTraceConfig(STAFString line)
{
    STAFCommandParser fTraceParser;

    //trace options
    fTraceParser.addOption("TRACE", 1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("ENABLE", 1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("DISABLE", 1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("ALL", 1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("TRACEPOINTS", 1,
        STAFCommandParser::kValueAllowed);
    fTraceParser.addOption("TRACEPOINT", 0,
        STAFCommandParser::kValueRequired);
    fTraceParser.addOption("SERVICES", 1,
        STAFCommandParser::kValueAllowed);
    fTraceParser.addOption("SERVICE", 0,
        STAFCommandParser::kValueRequired);
    fTraceParser.addOption("SET", 1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("DESTINATION", 1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("TO", 1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("STDOUT", 1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("STDERR", 1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("FILE", 1,
        STAFCommandParser::kValueRequired);
    fTraceParser.addOption("DEFAULTSERVICESTATE", 1,
        STAFCommandParser::kValueRequired);

    // trace groups
    fTraceParser.addOptionGroup("ENABLE DISABLE", 0, 1);
    fTraceParser.addOptionGroup("STDOUT STDERR FILE", 0, 1);
    fTraceParser.addOptionGroup("DEFAULTSERVICESTATE DESTINATION", 0, 1);

    // trace needs
    fTraceParser.addOptionNeed("ENABLE DISABLE",
        "ALL TRACEPOINT SERVICE TRACEPOINTS SERVICES");
    fTraceParser.addOptionNeed("SET", "TO DEFAULTSERVICESTATE");
    fTraceParser.addOptionNeed("DESTINATION", "SET");
    fTraceParser.addOptionNeed("TO", "DESTINATION");
    fTraceParser.addOptionNeed("TO", "STDOUT STDERR FILE");
    fTraceParser.addOptionNeed("STDOUT STDERR FILE DEFAULTSERVICESTATE", "SET");


    STAFCommandParseResultPtr parsedResult = fTraceParser.parse(line);

    if (parsedResult->rc != kSTAFOk)
    {
        cout << "Error on TRACE definition, " << line << endl
             << parsedResult->errorBuffer << endl;
        return 1;
    }

    STAFString errorBuffer;
    STAFString filename;
    STAFString tracepointList;
    STAFString serviceList;
    STAFRC_t rc = RESOLVE_OPTIONAL_STRING_OPTION("FILE", filename);

    if (rc != kSTAFOk)
    {
        cout << "Error on TRACE definition, " << line << endl
             << "Error with, " << errorBuffer << ", RC: " << rc << endl;
        return 1;
    }

    if (parsedResult->optionTimes("TO") != 0)
    {
        //set tracing information destination
        if (parsedResult->optionTimes("STDOUT") != 0)
            STAFTrace::setTraceDestination(kSTAFTraceToStdout);
        else if (parsedResult->optionTimes("STDERR") != 0)
            STAFTrace::setTraceDestination(kSTAFTraceToStderr);
        else
            STAFTrace::setTraceDestination(kSTAFTraceToFile, filename);
    }
    else if (parsedResult->optionTimes("DEFAULTSERVICESTATE") != 0)
    {
        //set default service tracing state
        bool stateToSetTo;
        STAFString defaultServiceState;

        rc = RESOLVE_STRING_OPTION("DEFAULTSERVICESTATE", defaultServiceState);

        if (rc != kSTAFOk)
        {
            cout << "Error on TRACE definition, " << line << endl
                 << "Error with, " << errorBuffer << ", RC: " << rc << endl;
            return 1;
        }

        defaultServiceState = defaultServiceState.upperCase();

        if(defaultServiceState == "DISABLED")
        {
            stateToSetTo = STAFServiceManager::kTraceDisabled;
        }
        else if (defaultServiceState == "ENABLED")
        {
            stateToSetTo = STAFServiceManager::kTraceEnabled;
        }
        else
        {
            cout << "Error on TRACE definition, " << line << endl
                 << "You must specify ENABLE or DISABLE"
                 << rc << endl;
            return kSTAFInvalidRequestString;
        }

        STAFServiceManager::setDefaultTraceState(stateToSetTo);
    }
    else
    {
        //set tracing of services and tracepoints
        bool stateToSetTo;

        if (parsedResult->optionTimes("ENABLE") != 0)
        {
            stateToSetTo = STAFServiceManager::kTraceEnabled;
        }
        else if (parsedResult->optionTimes("DISABLE") != 0)
        {
            stateToSetTo = STAFServiceManager::kTraceDisabled;
        }
        else
        {
            rc = kSTAFInvalidRequestString;
            cout << "Error on TRACE definition, " << line << endl
                 << "You must specify either ENABLE or DISABLE, RC: " << rc << endl;
            return rc;
        }

        if (parsedResult->optionTimes("ALL") != 0)
        //Handles ENABLE|DISABLE ALL ...
        {
            bool doTracepoints = false, doServices = false;

            if ((parsedResult->optionTimes("TRACEPOINTS") == 0) &&
                (parsedResult->optionTimes("SERVICES") == 0))
            {
                doTracepoints = true;
                doServices = true;
            }

            if (parsedResult->optionTimes("SERVICES") != 0)
                doServices = true;
            if (parsedResult->optionTimes("TRACEPOINTS") != 0)
                doTracepoints = true;

            if (doTracepoints)
            {
                if(stateToSetTo == STAFServiceManager::kTraceEnabled)
                       STAFTrace::traceOn(kSTAFTraceAll);
                else
                       STAFTrace::traceOff(kSTAFTraceAll);
            }
            if (doServices)
                STAFServiceManager::traceServicesChangeAll(stateToSetTo);
        }

        else
        {
            STAFString tracepointList = "";
            STAFTracePoint_t tracepointChangeSet = kSTAFTraceNone;

            if (parsedResult->optionTimes("TRACEPOINTS") != 0)
            {

                STAFString tracepoints;

                rc = RESOLVE_STRING_OPTION("TRACEPOINTS", tracepoints);

                if (rc != kSTAFOk)
                {
                    cout << "Error on TRACE definition, " << line << endl
                         << "Error with, " << errorBuffer << ", RC: "
                         << rc << endl;
                    return 1;
                }

                tracepointList = tracepoints;
            }

            if (parsedResult->optionTimes("TRACEPOINT") != 0)
            {
                STAFString tracepoint;
                for (int i = 1, optionCount =
                    parsedResult->optionTimes("TRACEPOINT"); i <= optionCount; ++i)
                {
                    rc = RESOLVE_INDEXED_STRING_OPTION("TRACEPOINT", i, tracepoint);

                    if (rc != kSTAFOk)
                    {
                        cout << "Error on TRACE definition, " << line << endl
                             << "Error with, " << errorBuffer << ", RC: "
                             << rc << endl;
                        return 1;
                    }

                    tracepointList += " " + tracepoint;
                }
            }

            tracepointList = tracepointList.upperCase();

            //validate list of tracepoints
            for (int i = 0; i < tracepointList.numWords(); ++i )
            {
                STAFString tracepoint = tracepointList.subWord(i, 1).upperCase();

                if (STAFTraceService::kSTAFTracepointNameMap.find(tracepoint) !=
                        STAFTraceService::kSTAFTracepointNameMap.end())
                {
                    tracepointChangeSet =
                        tracepointChangeSet |
                        STAFTraceService::kSTAFTracepointNameMap[tracepoint];
                }
                else
                {
                    cout << "Error on TRACE definition, " << line << endl
                         << "tracepoint " << tracepoint
                         << " does not exist, RC: " << kSTAFInvalidValue
                         << endl;
                    return 1;
                }
            }

            //now handle service tracing

            STAFString serviceList = "";

            if (parsedResult->optionTimes("SERVICES") != 0)
            {
                STAFString services;

                rc = RESOLVE_STRING_OPTION("SERVICES", services);

                if (rc != kSTAFOk)
                {
                    cout << "Error on TRACE definition, " << line << endl
                         << "Error with, " << errorBuffer << ", RC: " << rc << endl;
                    return 1;
                }

                serviceList = services;
            }

            if (parsedResult->optionTimes("SERVICE") != 0)
            {
                STAFString service;

                for (int i = 1, optionCount =
                    parsedResult->optionTimes("SERVICE"); i<= optionCount; ++i)
                {

                    rc = RESOLVE_INDEXED_STRING_OPTION("SERVICE", i, service);
                    serviceList += " " + service;

                    if (rc != kSTAFOk)
                    {
                        cout << "Error on TRACE definition, " << line << endl
                             << "Error with, " << errorBuffer << ", RC: " << rc << endl;
                        return 1;
                    }
                }
            }

            //after ensuring there are no errors, make desired changes
            serviceList = serviceList.upperCase();

            STAFServiceManager::traceServicesChange(serviceList, stateToSetTo);

            if(stateToSetTo == STAFServiceManager::kTraceEnabled)
                STAFTrace::traceOn(tracepointChangeSet);
            else
                STAFTrace::traceOff(tracepointChangeSet);
        }
    }
    return kSTAFOk;
}

unsigned int handleServiceConfig(STAFString line)
{
    STAFCommandParser fServiceParser;

    // SERVICE <Name>
    // LIBRARY <Library> [EXECUTE <Exec>] {OPTION <OpName[=OpValue]>}
    //     [PARMS <Parms>]
    // or
    // DELEGATE <Machine> [TONAME <Service Name>]

    // service config options

    fServiceParser.addOption("SERVICE",  1,
        STAFCommandParser::kValueRequired);
    fServiceParser.addOption("LIBRARY",    1,
        STAFCommandParser::kValueRequired);
    fServiceParser.addOption("EXECUTE",    1,
        STAFCommandParser::kValueRequired);
    fServiceParser.addOption("OPTION",    0,
        STAFCommandParser::kValueRequired);
    fServiceParser.addOption("PARMS",    1,
        STAFCommandParser::kValueRequired);
    fServiceParser.addOption("DELEGATE", 1,
        STAFCommandParser::kValueRequired);
    fServiceParser.addOption("TONAME",   1,
        STAFCommandParser::kValueRequired);

    // service config groups

    fServiceParser.addOptionGroup("SERVICE",          1, 1);
    fServiceParser.addOptionGroup("LIBRARY DELEGATE", 1, 1);

    // service config needs

    fServiceParser.addOptionNeed("EXECUTE OPTION PARMS",  "LIBRARY");
    fServiceParser.addOptionNeed("TONAME", "DELEGATE");

    STAFCommandParseResultPtr parsedResult = fServiceParser.parse(line);

    if (parsedResult->rc != kSTAFOk)
    {
        cout << "Error on Service definition, " << line << endl
             << parsedResult->errorBuffer << endl;
        return 1;
    }

    STAFString errorBuffer;
    STAFString name;
    STAFRC_t rc = RESOLVE_STRING_OPTION("SERVICE", name);

    if (rc)
    {
        cout << "Error on Service definition, " << line << endl
             << "Error resolving " << parsedResult->optionValue("SERVICE")
             << ", RC: " << rc << endl;
        return 1;
    }

    rc = gServiceManagerPtr->verifyServiceName(name);

    if (rc)
    {
        cout << "Error on Service definition, " << line << endl
             << "Service name " << name << " is not valid, RC: " << rc << endl;
        cout << INVALID_SERVICE_NAME_ERROR_MESSAGE() << endl;

        return 1;
    }

    try
    {
        STAFServicePtr service;

        if (parsedResult->optionTimes("LIBRARY") != 0)
        {
            // Setup external service data
            STAFString library;
            STAFString exec;
            STAFString parms;
            STAFExternalService::OptionList options;

            rc = RESOLVE_STRING_OPTION("LIBRARY", library);

            if (!rc) rc = RESOLVE_STRING_OPTION("EXECUTE", exec);
            if (!rc) rc = RESOLVE_STRING_OPTION("PARMS", parms);

            for (int i = 1; !rc && (i <= parsedResult->optionTimes("OPTION"));
                 ++i)
            {
                STAFString option;
                rc = RESOLVE_INDEXED_STRING_OPTION("OPTION", i, option);
                options.push_back(option);
            }

            if (rc)
            {
                cout << "Error on Service definition, " << line << endl
                     << "Error with " << errorBuffer << ", RC: " << rc
                     << endl;
                return 1;
            }

            service = STAFServicePtr(new STAFExternalService(name, library,
                                     exec, options, parms,
                                     kSTAFServiceTypeService),
                                     STAFServicePtr::INIT);
        }
        else
        {
            // This is a Delegated service

            STAFString toMachine;
            STAFString toName = name;

            rc = RESOLVE_STRING_OPTION("DELEGATE", toMachine);

            if (!rc) RESOLVE_OPTIONAL_STRING_OPTION("TONAME", toName);

            if (rc)
            {
                cout << "Error on Service definition, " << line << endl
                     << "Error with " << errorBuffer << ", RC: " << rc << endl;
                return 1;
            }

            service = STAFServicePtr(new STAFDelegatedService(name, toMachine,
                                     toName), STAFServicePtr::INIT);
        }

        rc = gServiceManagerPtr->add(service);

        if (rc)
        {
            cout << "Error, RC: " << rc << ", adding Service definition, "
                 << line << endl;
            return 1;
        }
    }
    catch (STAFServiceCreateException &e)
    {
        cout << "Error on Service definition, " << line << endl
             << e.getErrorCode() << ":" << e.getText() << endl;
        return 1;
    }
    catch (STAFException &e)
    {
        e.write();
        return 1;
    }
    catch (...)
    {
        // XXX: Why is this here?
        cout << "Exception caught. Need to get this out (STAFConfig.cpp)"
             << endl;
        return 1;
    }

    return 0;
}


unsigned int handleServiceLoaderConfig(STAFString line)
{
    STAFCommandParser fServiceLoaderParser;

    // SERVICELOADER
    // LIBRARY <Library> [EXECUTE <Exec>] {OPTION <OpName[=OpValue]>}
    //     [PARMS <Parms>]

    // serviceloader config options

    fServiceLoaderParser.addOption("SERVICELOADER",  1,
        STAFCommandParser::kValueNotAllowed);
    fServiceLoaderParser.addOption("LIBRARY",    1,
        STAFCommandParser::kValueRequired);
    fServiceLoaderParser.addOption("EXECUTE",    1,
        STAFCommandParser::kValueRequired);
    fServiceLoaderParser.addOption("OPTION",    0,
        STAFCommandParser::kValueRequired);
    fServiceLoaderParser.addOption("PARMS",    1,
        STAFCommandParser::kValueRequired);

    // service config groups

    fServiceLoaderParser.addOptionGroup("SERVICELOADER", 1, 1);
    fServiceLoaderParser.addOptionGroup("LIBRARY", 1, 1);

    // service config needs

    fServiceLoaderParser.addOptionNeed("EXECUTE OPTION PARMS",  "LIBRARY");

    STAFCommandParseResultPtr parsedResult = fServiceLoaderParser.parse(line);

    if (parsedResult->rc != kSTAFOk)
    {
        cout << "Error on ServiceLoader definition, " << line << endl
             << parsedResult->errorBuffer << endl;
        return 1;
    }

    try
    {
        STAFServicePtr service;

        // Setup external service data
        STAFString library;
        STAFString exec;
        STAFString parms;
        STAFExternalService::OptionList options;
        STAFString errorBuffer;
        STAFRC_t rc = RESOLVE_STRING_OPTION("LIBRARY", library);

        if (!rc) rc = RESOLVE_STRING_OPTION("EXECUTE", exec);
        if (!rc) rc = RESOLVE_STRING_OPTION("PARMS", parms);

        for (int i = 1; !rc && (i <= parsedResult->optionTimes("OPTION"));
             ++i)
        {
            STAFString option;
            rc = RESOLVE_INDEXED_STRING_OPTION("OPTION", i, option);
            options.push_back(option);
        }

        if (rc)
        {
            cout << "Error on ServiceLoader definition, " << line << endl
                 << "Error with " << errorBuffer << ", RC: " << rc
                 << endl;
            return 1;
        }

        service = STAFServicePtr(new STAFExternalService("STAFServiceLoader" +
                                                         STAFString(fSLSNum),
                                library, exec, options, parms,
                                kSTAFServiceTypeServiceLoader),
                                STAFServicePtr::INIT);

        fSLSNum++;

        rc = gServiceManagerPtr->addSLS(service);

        if (rc)
        {
            cout << "Error, RC: " << rc << ", adding ServiceLoader definition, "
                 << line << endl;
            return 1;
        }
    }
    catch (STAFServiceCreateException &e)
    {
        cout << "Error on ServiceLoader definition, " << line << endl
             << e.getErrorCode() << ":" << e.getText() << endl;
        return 1;
    }
    catch (STAFException &e)
    {
        e.write();
        return 1;
    }
    catch (...)
    {
        // XXX: Why is this here?
        cout << "Exception caught. Need to get this out (STAFConfig.cpp)"
             << endl;
        return 1;
    }

    return 0;
}


unsigned int handleAuthenticatorConfig(STAFString line)
{
    STAFCommandParser fAuthenticatorParser;

    // AUTHENTICATOR <Name>
    // LIBRARY <Library> [EXECUTE <Exec>] {OPTION <OpName[=OpValue]>}
    //     [PARMS <Parms>]

    // Authenticator config options

    fAuthenticatorParser.addOption("AUTHENTICATOR",  1,
        STAFCommandParser::kValueRequired);
    fAuthenticatorParser.addOption("LIBRARY",    1,
        STAFCommandParser::kValueRequired);
    fAuthenticatorParser.addOption("EXECUTE",    1,
        STAFCommandParser::kValueRequired);
    fAuthenticatorParser.addOption("OPTION",    0,
        STAFCommandParser::kValueRequired);
    fAuthenticatorParser.addOption("PARMS",    1,
        STAFCommandParser::kValueRequired);

    // service config groups

    fAuthenticatorParser.addOptionGroup("AUTHENTICATOR", 1, 1);
    fAuthenticatorParser.addOptionGroup("LIBRARY", 1, 1);

    // service config needs

    fAuthenticatorParser.addOptionNeed("EXECUTE OPTION PARMS", "LIBRARY");

    STAFCommandParseResultPtr parsedResult = fAuthenticatorParser.parse(line);

    if (parsedResult->rc != kSTAFOk)
    {
        cout << "Error on Authenticator definition, " << line << endl
             << parsedResult->errorBuffer << endl;
        return 1;
    }

    STAFString errorBuffer;
    STAFString name;
    STAFRC_t rc = RESOLVE_STRING_OPTION("AUTHENTICATOR", name);

    if (rc)
    {
        cout << "Error on Authenticator definition, " << line << endl
             << "Error resolving " << parsedResult->optionValue("AUTHENTICATOR")
             << ", RC: " << rc << endl;
        return 1;
    }

    rc = gServiceManagerPtr->verifyServiceName(name);

    if (rc != kSTAFOk)
    {
        cout << "Error on Service definition, " << line << endl
             << "Service name " << name << " is not valid, RC: " << rc << endl;
        cout << INVALID_SERVICE_NAME_ERROR_MESSAGE() << endl;
        return 1;
    }

    try
    {
        STAFServicePtr service;

        // Setup external service data
        STAFString library;
        STAFString exec;
        STAFString parms;
        STAFExternalService::OptionList options;

        rc = RESOLVE_STRING_OPTION("LIBRARY", library);

        if (!rc) rc = RESOLVE_STRING_OPTION("EXECUTE", exec);
        if (!rc) rc = RESOLVE_STRING_OPTION("PARMS", parms);

        for (int i = 1; !rc && (i <= parsedResult->optionTimes("OPTION"));
             ++i)
        {
            STAFString option;
            rc = RESOLVE_INDEXED_STRING_OPTION("OPTION", i, option);
            options.push_back(option);
        }


        if (rc)
        {
            cout << "Error on Authenticator definition, " << line << endl
                 << "Error with " << errorBuffer << ", RC: " << rc
                 << endl;
            return 1;
        }

        service = STAFServicePtr(new STAFExternalService(name, library,
                                 exec, options, parms,
                                 kSTAFServiceTypeAuthenticator),
                                 STAFServicePtr::INIT);

        rc = gServiceManagerPtr->addAuthenticator(service);

        if (rc)
        {
            cout << "Error, RC: " << rc
                << ", adding Authenticator definition, "
                << line << endl;
            return 1;
        }

        // Assign as default authenticator if it has not been specified yet.
        if ((gServiceManagerPtr->getDefaultAuthenticator() == gNoneString) ||
            (gServiceManagerPtr->getDefaultAuthenticator() == ""))
        {
            gServiceManagerPtr->setDefaultAuthenticator(name);
        }
    }
    catch (STAFServiceCreateException &e)
    {
        cout << "Error on Authenticator definition, " << line << endl
             << e.getErrorCode() << ":" << e.getText() << endl;
        return 1;
    }
    catch (STAFException &e)
    {
        e.write();
        return 1;
    }
    catch (...)
    {
        // XXX: Why is this here?
        cout << "Exception caught. Need to get this out (STAFConfig.cpp)"
             << endl;
        return 1;
    }

    return 0;
}


unsigned int handleSetConfig(STAFString line)
{
    STAFCommandParser fSetParser;

    // set config options

    // XXX: Note ALLOWMULTIREG is deprecated.  It is currently allowed by the
    //      parser, but ignored everywhere else, for backward compatibility.
    //      It will be removed entirely in STAF V3.0

    fSetParser.addOption("SET",                   1,
        STAFCommandParser::kValueNotAllowed);
    fSetParser.addOption("SYSTEM",                1,
        STAFCommandParser::kValueNotAllowed);
    fSetParser.addOption("SHARED",                1,
        STAFCommandParser::kValueNotAllowed);
    fSetParser.addOption("VAR",                   0,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("DATADIR",               1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("CONNECTATTEMPTS",       1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("CONNECTRETRYDELAY",     1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("INTERFACECYCLING",      1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("MAXFILES",              1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("MAXQUEUESIZE",          1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("INITIALTHREADS",        1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("THREADGROWTHDELTA",     1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("ALLOWMULTIREG",         1,
        STAFCommandParser::kValueNotAllowed);
    fSetParser.addOption("DEFAULTINTERFACE",      1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("DEFAULTAUTHENTICATOR",  1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("DEFAULTSTOPUSING",      1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("DEFAULTNEWCONSOLE",     1,
        STAFCommandParser::kValueNotAllowed);
    fSetParser.addOption("DEFAULTSAMECONSOLE",    1,
        STAFCommandParser::kValueNotAllowed);
    fSetParser.addOption("DEFAULTAUTHDISABLEDACTION", 1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("DEFAULTAUTHUSERNAME",   1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("DEFAULTAUTHPASSWORD",   1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("PROCESSAUTHMODE",       1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("DEFAULTSHELL",          1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("DEFAULTNEWCONSOLESHELL", 1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("DEFAULTSAMECONSOLESHELL", 1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("ENABLEDIAGS", 1,
        STAFCommandParser::kValueNotAllowed);
    fSetParser.addOption("STRICTFSCOPYTRUST", 1,
        STAFCommandParser::kValueNotAllowed);
    fSetParser.addOption("RESULTCOMPATIBILITYMODE", 1,
        STAFCommandParser::kValueRequired);

    // set config groups

    fSetParser.addOptionGroup("SET", 1 , 1);
    fSetParser.addOptionGroup("DEFAULTNEWCONSOLE DEFAULTSAMECONSOLE", 0 , 1);
    fSetParser.addOptionGroup("SYSTEM SHARED", 0, 1);
    fSetParser.addOptionNeed("SHARED SYSTEM", "VAR");

    STAFCommandParseResultPtr parsedResult = fSetParser.parse(line);

    if (parsedResult->rc != kSTAFOk)
    {
        cout << "Error on SET statement, " << line << endl
             << parsedResult->errorBuffer << endl;
        return 1;
    }

    STAFString errorBuffer;

    if (parsedResult->optionTimes("VAR") != 0)
    {
        // handle var config

        STAFVariablePoolPtr thePool;

        if (parsedResult->optionTimes("SHARED") != 0)
        {
            thePool = *gSharedVariablePoolPtr;
        }
        else
        {
            thePool = *gGlobalVariablePoolPtr;
        }

        for(int i = 1; i <= parsedResult->optionTimes("VAR"); ++i)
        {
            STAFString nameAndValue = parsedResult->optionValue("VAR", i);
            unsigned int equalPos = nameAndValue.find(kUTF8_EQUAL);

            if (equalPos == STAFString::kNPos)
            {
                cout << "Error on variable definition, " << line << endl
                     << errorBuffer << endl;
                return 1;
            }

            thePool->set(nameAndValue.subString(0, equalPos),
                         nameAndValue.subString(equalPos +
                         nameAndValue.sizeOfChar(equalPos)));
        }

        return 0;
    }

    STAFRC_t rc = kSTAFOk;

    if (parsedResult->optionTimes("DATADIR") != 0)
    {
        // Resolve the value for the DATADIR operational setting

        STAFString dataDir;

        rc = RESOLVE_STRING_OPTION("DATADIR", dataDir);

        if (rc != kSTAFOk)
        {
            cout << "Error on DATADIR setting, " << line << endl
                 << "Error resolving "
                 << parsedResult->optionValue("DATADIR") << ", RC: "
                 << rc << endl;
            return 1;
        }

        // Create the data directory if it doesn't exist

        STAFFSPath dataPath;
        dataPath.setRoot(dataDir);

        if (!dataPath.exists())
        {
            try
            {
                dataPath.createDirectory(kSTAFFSCreatePath);
            }
            catch (...)
            { /* Do Nothing */ }

            if (!dataPath.exists())
            {
                cout << "Error creating DATADIR directory: "
                     << dataDir << endl;
                return 1;
            }
        }

        *gSTAFWriteLocationPtr = dataDir;

        // Set STAF system variable STAF/DataDir to the datadir value
        // This is a "read-only" STAF variable.

        (*gGlobalVariablePoolPtr)->set("STAF/DataDir", dataDir);
    }

    if (parsedResult->optionTimes("CONNECTATTEMPTS") != 0)
    {
        // Handle CONNECTATTEMPTS

        STAFString unresConnectAttempts =
            parsedResult->optionValue("CONNECTATTEMPTS");

        STAFString connectAttempts;

        rc = RESOLVE_STRING_OPTION("CONNECTATTEMPTS", connectAttempts);

        if (rc != kSTAFOk)
        {
            cout << "Error on CONNECTATTEMPTS setting, " << line << endl
                 << "Error resolving "
                 << parsedResult->optionValue("CONNECTATTEMPTS") << ", RC: "
                 << rc << endl;
            return 1;
        }

        if (!connectAttempts.isDigits())
        {
            cout << "Error on CONNECTATTEMPTS setting, " << line << endl
                 << "CONNECTATTEMPTS must be a positive integer" << endl;
            return 1;
        }

        gConnectionAttempts = connectAttempts.asUInt();

        if (gConnectionAttempts < 1)
        {
            cout << "Error on CONNECTATTEMPTS setting, " << line << endl
                 << "CONNECTATTEMPTS must be a positive integer" << endl;
            return 1;
        }
    }

    if (parsedResult->optionTimes("CONNECTRETRYDELAY") != 0)
    {
        // Handle CONNECTRETRYDELAY

        STAFString unresConnectRetryDelay =
            parsedResult->optionValue("CONNECTRETRYDELAY");

        STAFString connectRetryDelay;

        rc = RESOLVE_STRING_OPTION("CONNECTRETRYDELAY", connectRetryDelay);

        if (rc != kSTAFOk)
        {
            cout << "Error on CONNECTRETRYDELAY setting, " << line << endl
                 << "Error resolving "
                 << parsedResult->optionValue("CONNECTRETRYDELAY") << ", RC: "
                 << rc << endl;
            return 1;
        }

        if (!connectRetryDelay.isDigits())
        {
            cout << "Error on CONNECTRETRYDELAY setting, " << line << endl
                 << "CONNECTRETRYDELAY must be a non-negative integer" << endl;
            return 1;
        }

        gConnectionRetryDelay = connectRetryDelay.asUInt();
    }
    
    if (parsedResult->optionTimes("INTERFACECYCLING") != 0)
    {
        // Handle INTERFACECYCLING

        STAFString unresInterfaceCycling = parsedResult->optionValue(
            "INTERFACECYCLING");

        STAFString interfaceCycling;

        rc = RESOLVE_STRING_OPTION("INTERFACECYCLING", interfaceCycling);

        if (rc != kSTAFOk)
        {
            cout << "Error on INTERFACECYCLING setting, " << line << endl
                 << "Error resolving " 
                 << parsedResult->optionValue("INTERFACECYCLING") << ", RC: "
                 << rc << endl;
            return 1;
        }

        if (interfaceCycling.toUpperCase() == "ENABLED")
        {
            gConnectionManagerPtr->enableAutoInterfaceCycling();
        }
        else if (interfaceCycling.toUpperCase() == "DISABLED")
        {
            gConnectionManagerPtr->disableAutoInterfaceCycling();
        }
        else
        {
            cout << "Error on INTERFACECYCLING setting, " << line << endl
                 << "INTERFACECYCLING must be set to Enabled or Disabled"
                 << endl;
            return 1;
        }
    }
    
    if (parsedResult->optionTimes("MAXFILES") != 0)
    {
        // Handle MAXFILES

        STAFString maxFiles;

        rc = RESOLVE_STRING_OPTION("MAXFILES", maxFiles);

        if (rc != kSTAFOk)
        {
            cout << "Error on MAXFILES setting, " << line << endl
                 << "Error resolving " << parsedResult->optionValue("MAXFILES")
                 << ", RC: " << rc << endl;
            return 1;
        }

        if (!maxFiles.isDigits())
        {
            cout << "Error on MAXFILES setting, " << line << endl
                 << "MAXFILES must be a non-negative integer" << endl;
            return 1;
        }

        gMaxFiles = (long)maxFiles.asUInt();
    }

    if (parsedResult->optionTimes("MAXQUEUESIZE") != 0)
    {
        // Handle MAXQUEUESIZE

        STAFString maxQueueSize;

        rc = RESOLVE_STRING_OPTION("MAXQUEUESIZE", maxQueueSize);

        if (rc != kSTAFOk)
        {
            cout << "Error on MAXQUEUESIZE setting, " << line << endl
                 << "Error resolving "
                 << parsedResult->optionValue("MAXQUEUESIZE") << ", RC: "
                 << rc << endl;
            return 1;
        }

        if (!maxQueueSize.isDigits())
        {
            cout << "Error on MAXQUEUESIZE setting, " << line << endl
                 << "MAXQUEUESIZE must be a non-negative integer" << endl;
            return 1;
        }

        gMaxQueueSize = maxQueueSize.asUInt();
    }

    if (parsedResult->optionTimes("INITIALTHREADS") != 0)
    {
        // Handle INITIALTHREADS

        STAFString initialThreads;

        rc = RESOLVE_STRING_OPTION("INITIALTHREADS", initialThreads);

        if (rc != kSTAFOk)
        {
            cout << "Error on INITIALTHREADS setting, " << line  << endl
                 << "Error resolving "
                 << parsedResult->optionValue("INITIALTHREADS") << ", RC: "
                 << rc << endl;
            return 1;
        }

        if (!initialThreads.isDigits() ||
            (initialThreads.asUInt() < 1))
        {
            cout << "Error on INITIALTHREADS setting, " << line << endl
                 << "INITIALTHREADS must be a positive integer" << endl;
            return 1;
        }

        gNumInitialThreads = initialThreads.asUInt();
    }

    if (parsedResult->optionTimes("THREADGROWTHDELTA") != 0)
    {
        // Handle THREADGROWTHDELTA

        STAFString threadDelta;

        rc = RESOLVE_STRING_OPTION("THREADGROWTHDELTA", threadDelta);

        if (rc != kSTAFOk)
        {
            cout << "Error on THREADGROWTHDELTA setting, " << line << endl
                 << "Error resolving "
                 << parsedResult->optionValue("THREADGROWTHDELTA") << ", RC: "
                 << rc << endl;
            return 1;
        }

        if (!threadDelta.isDigits() || (threadDelta.asUInt() < 1))
        {
            cout << "Error on THREADGROWTHDELTA setting, " << line << endl
                 << "THREADGROWTHDELTA must be a positive integer" << endl;
            return 1;
        }

        gThreadManagerPtr->setGrowthDelta(threadDelta.asUInt());
    }

    if (parsedResult->optionTimes("DEFAULTINTERFACE") != 0)
    {
        STAFString defaultInterface;

        rc = RESOLVE_STRING_OPTION("DEFAULTINTERFACE", defaultInterface);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTINTERFACE setting, " << line << endl
                 << "Error resolving "
                 << parsedResult->optionValue("DEFAULTINTERFACE") << ", RC: "
                 << rc << endl;
            return 1;
        }

        rc = gConnectionManagerPtr->setDefaultConnectionProvider(
            defaultInterface);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTINTERFACE setting, " << line << endl
                 << "Interface " << defaultInterface
                 << " is not registered, RC: " << rc << endl;
            return 1;
        }
    }

    if (parsedResult->optionTimes("DEFAULTAUTHENTICATOR") != 0)
    {
        STAFString defaultAuth;

        rc = RESOLVE_STRING_OPTION("DEFAULTAUTHENTICATOR", defaultAuth);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTAUTHENTICATOR setting, " << line << endl
                 << "Error resolving "
                 << parsedResult->optionValue("DEFAULTAUTHENTICATOR") << ", RC: "
                 << rc << endl;
            return 1;
        }

        rc = gServiceManagerPtr->setDefaultAuthenticator(defaultAuth);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTAUTHENTICATOR setting, " << line << endl
                 << "Authenticator " << defaultAuth << " is not registered, RC: "
                 << rc << endl;
            return 1;
        }
    }

    if (parsedResult->optionTimes("DEFAULTSTOPUSING") != 0)
    {
        STAFString stopString;
        STAFProcessStopMethod_t stopMethod;

        rc = RESOLVE_STRING_OPTION("DEFAULTSTOPUSING", stopString);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTSTOPUSING setting, " << line << endl
                 << "Error resolving "
                 << parsedResult->optionValue("DEFAULTSTOPUSING") << ", RC: "
                 << rc << endl;
            return 1;
        }

        rc = STAFProcessService::getStopMethodFromString(stopMethod,
                                                         stopString);
        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTSTOPUSING setting, " << line << endl
                 << "Invalid value, " << stopString << ", RC: "
                 << rc << endl;
            return 1;
        }

        rc = STAFProcessService::setDefaultStopMethod(stopMethod);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTSTOPUSING setting, " << line << endl
                 << "Error changing the default, RC: " << rc << endl;
            return 1;
        }
    }

    if (parsedResult->optionTimes("PROCESSAUTHMODE") != 0)
    {
        STAFString authModeString;
        STAFProcessAuthenticationMode_t authMode;

        rc = RESOLVE_STRING_OPTION("PROCESSAUTHMODE", authModeString);

        if (rc != kSTAFOk)
        {
            cout << "Error on PROCESSAUTHMODE setting, " << line << endl
                 << "Error resolving "
                 << parsedResult->optionValue("PROCESSAUTHMODE") << ", RC: "
                 << rc << endl;
            return 1;
        }

        rc = STAFProcessService::getAuthModeFromString(authMode,
                                                       authModeString);

        if (rc != kSTAFOk)
        {
            cout << "Error on PROCESSAUTHMODE setting, " << line << endl
                 << "Invalid value, " << authModeString << ", RC: "
                 << rc << endl;
            return 1;
        }

        unsigned int osRC = 0;

        rc = STAFProcessService::setAuthMode(authMode, osRC);

        if (rc != kSTAFOk)
        {
            cout << "Error on PROCESSAUTHMODE setting, " << line << endl
                 << "Error changing the default, RC: " << rc << endl;
            return 1;
        }
    }

    if (parsedResult->optionTimes("DEFAULTAUTHUSERNAME") != 0)
    {
        STAFString authUsernameString;

        rc = RESOLVE_STRING_OPTION("DEFAULTAUTHUSERNAME", authUsernameString);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTAUTHUSERNAME setting, " << line << endl
                 << "Error resolving "
                 << parsedResult->optionValue("DEFAULTAUTHUSERNAME") << ", RC: "
                 << rc << endl;
            return 1;
        }

        rc = STAFProcessService::setDefaultAuthUsername(authUsernameString);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTAUTHUSERNAME setting, " << line << endl
                 << "Error changing the default, RC: " << rc << endl;
            return 1;
        }
    }

    if (parsedResult->optionTimes("DEFAULTAUTHPASSWORD") != 0)
    {
        STAFString authPasswordString;

        rc = RESOLVE_STRING_OPTION("DEFAULTAUTHPASSWORD", authPasswordString);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTAUTHPASSWORD setting, " << line << endl
                 << "Error resolving "
                 << parsedResult->optionValue("DEFAULTAUTHPASSWORD") << ", RC: "
                 << rc << endl;
            return 1;
        }

        rc = STAFProcessService::setDefaultAuthPassword(authPasswordString);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTAUTHPASSWORD setting, " << line << endl
                 << "Error changing the default, RC: " << rc << endl;
            return 1;
        }
    }

    if (parsedResult->optionTimes("DEFAULTAUTHDISABLEDACTION") != 0)
    {
        STAFString defaultAuthDisabledActionString;
        STAFProcessDisabledAuthAction_t defaultAuthDisabledAction;

        rc = RESOLVE_STRING_OPTION("DEFAULTAUTHDISABLEDACTION",
                                   defaultAuthDisabledActionString);
        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTAUTHDISABLEDACTION setting, " << line
                 << endl << "Error resolving "
                 << parsedResult->optionValue("DEFAULTAUTHDISABLEDACTION")
                 << ", RC: " << rc << endl;
            return 1;
        }

        rc = STAFProcessService::getDefaultDisabledAuthActionFromString(
             defaultAuthDisabledAction, defaultAuthDisabledActionString);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTAUTHDISABLEDACTION setting, " << line
                 << endl
                 << "Invalid value, " << defaultAuthDisabledActionString
                 << ", RC: "
                 << rc << endl;
            return 1;
        }

        rc = STAFProcessService::setDefaultDisabledAuthAction(
             defaultAuthDisabledAction);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTAUTHDISABLEDACTION setting, " << line
                 << endl
                 << "Error changing the default, RC: " << rc << endl;
            return 1;
        }
    }

    if (parsedResult->optionTimes("DEFAULTNEWCONSOLE") != 0)
        gDefaultConsoleMode = kSTAFProcessNewConsole;
    else if (parsedResult->optionTimes("DEFAULTSAMECONSOLE") != 0)
        gDefaultConsoleMode = kSTAFProcessSameConsole;

    if (parsedResult->optionTimes("DEFAULTSHELL") != 0)
    {
        STAFString shellCommand;

        rc = RESOLVE_STRING_OPTION("DEFAULTSHELL", shellCommand);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTSHELL setting, " << line << endl
                 << "Error resolving "
                 << parsedResult->optionValue("DEFAULTSHELL") << ", RC: "
                 << rc << endl;
            return 1;
        }

        rc = STAFProcessService::setDefaultShellCommand(shellCommand);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTSHELL setting, " << line << endl
                 << "Error changing the default, RC: " << rc << endl;
            return 1;
        }
    }

    if (parsedResult->optionTimes("DEFAULTSAMECONSOLESHELL") != 0)
    {
        STAFString shellCommand;

        rc = RESOLVE_STRING_OPTION("DEFAULTSAMECONSOLESHELL", shellCommand);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTSAMECONSOLESHELL setting, " << line
                 << endl << "Error resolving "
                 << parsedResult->optionValue("DEFAULTSAMECONSOLESHELL")
                 << ", RC: " << rc << endl;
            return 1;
        }

        rc = STAFProcessService::setDefaultSameConsoleShell(shellCommand);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTSAMECONSOLESHELL setting, " << line
                 << endl << "Error changing the default, RC: " << rc << endl;
            return 1;
        }
    }

    if (parsedResult->optionTimes("DEFAULTNEWCONSOLESHELL") != 0)
    {
        STAFString shellCommand;

        rc = RESOLVE_STRING_OPTION("DEFAULTNEWCONSOLESHELL", shellCommand);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTNEWCONSOLESHELL setting, " << line
                 << endl << "Error resolving "
                 << parsedResult->optionValue("DEFAULTNEWCONSOLESHELL")
                 << ", RC: " << rc << endl;
            return 1;
        }

        rc = STAFProcessService::setDefaultNewConsoleShell(shellCommand);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTNEWCONSOLESHELL setting, " << line
                 << endl << "Error changing the default, RC: " << rc << endl;
            return 1;
        }
    }

    // Handle ENABLEDIAGS

    if (parsedResult->optionTimes("ENABLEDIAGS") != 0)
    {
        rc = gDiagManagerPtr->enable();

        if (rc != kSTAFOk)
        {
            cout << "Error on ENABLEDIAGS setting, " << line
                 << endl << "RC: " << rc << endl;
            return 1;
        }
    }

    // Handle STRICTFSCOPYTRUST

    if (parsedResult->optionTimes("STRICTFSCOPYTRUST") != 0)
    {
        gStrictFSCopyTrust = 1;
    }
    
    // Handle RESULTCOMPATIBILITYMODE

    if (parsedResult->optionTimes("RESULTCOMPATIBILITYMODE") != 0)
    {
        STAFString mode;

        rc = RESOLVE_STRING_OPTION("RESULTCOMPATIBILITYMODE", mode);

        if (rc != kSTAFOk)
        {
            cout << "Error on RESULTCOMPATIBILITYMODE setting, " << line
                 << endl << "Error resolving "
                 << parsedResult->optionValue("RESULTCOMPATIBILITYMODE")
                 << ", RC: " << rc << endl;
            return 1;
        }

        if (mode.isEqualTo("NONE", kSTAFStringCaseInsensitive))
        {
            gResultCompatibilityMode = kSTAFResultCompatibilityNone;
        }
        else if (mode.isEqualTo("VERBOSE", kSTAFStringCaseInsensitive))
        {
            gResultCompatibilityMode = kSTAFResultCompatibilityVerbose;
        }
        else
        {
            cout << "Error on RESULTCOMPATIBILITYMODE setting, " << line
                 << endl << "Invalid value.  Must be None or Verbose,  RC: "
                 << kSTAFInvalidValue << endl;

            return 1;
        }
    }

    return 0;
}
